"""
Alpaca has functionality to embed semantic information provided by ontologies
by reading annotations that were inserted into Python objects. The annotations
are used by the :class:`alpaca.AlpacaProvDocument` class when serializing the
provenance information as RDF. The annotations will be inserted as additional
`rdf:type` triples alongside the classes already defined by the Alpaca PROV
model.

It expects that the Python object has a dictionary stored as the special
`__ontology__` attribute. All the specific annotations for the Python object
are contained in this dictionary. The annotations are URIs of the
relevant ontology classes that represent the Python object (e.g., a function)
or one of its elements (e.g., the parameters of the function). If providing a
full URI (i.e., without an ontology namespace as prefix - CURIEs), the URI
must start with `<` and end with `>` (e.g.,
`<http://example.org/ontology#Class>`). CURIEs are allowed in the form
`ontology:Class`.

Currently, annotations for two main Python objects are supported: functions
(intended for a Python function object) and data objects (intended for
instances of objects instantiated from a Python class). Specific keys in the
`__ontology__` dictionary will define the main URI describing either the
function or the data object:

* 'function' : str or list of str
   A URI to the ontology class representing the Python function. Multiple URIs
   can be passed as a list, if the function is represented by multiple classes.
* 'data_object' : str or list of str
   A URI to the ontology class representing the Python data object. Multiple
   URIs can be passed as a list, if the object is represented by multiple
   classes.

Additional annotations can be stored depending on whether a function or data
object is being annotated.

For functions, the additional items that can be stored in the `__ontology__`
dictionary are:

* 'arguments' : dict
   A dictionary where the keys are argument names (cf. the function
   declaration in the `def` statement) and the values are the URI(s)
   to the ontology class(es) representing the argument.
* 'returns' : dict
   A dictionary where the keys are function outputs, and the values define the
   URI(s) to the ontology class(es) representing each output identified by a
   key.
   The keys in the `returns` dictionary can have three possible values:
   1. a string with one output name (if this is the name of an argument, cf.
   the function declaration in the `def` statement), which assumes that a
   function uses one of the arguments as output, or
   2. an integer corresponding to the order of the output (as defined by the
   function `return` statements), or
   3. a string of `*`s. In this case, the function returns a container (e.g.,
   a list, or list of lists), and the number of `*`s defines the depth
   within the container whose elements are defined by the ontology class in
   the value. For instance, if the function returns a list of lists
   `[[obj1, obj2], [obj3, obj4], ...]`, the objects are at the third level. If
   each element `objX` is represented by an ontology class defined by `<URI>`,
   one can annotate the elements with a key `'***'` and value `'<URI>'`. An
   annotation with key `'*'` would be used to annotate the main list returned,
   and the key `'**'` would be used to annotate each inner list inside the
   main list (e.g., `[obj1, obj2]` or `[obj3, obj4]`). This is useful for using
   ontologies that define concepts that represent groups of elements.

For data objects, the additional items that can be stored in the `__ontology__`
dictionary are:

* 'attributes' : dict
   A dictionary where the keys are object attribute names and the values are
   the URI(s) to the ontology class(es) representing the attribute.
* 'annotations' : dict
   A dictionary where the keys are annotation names and the values are the
   URI(s) to the ontology class(es) representing the annotation. Annotations
   are key-pair values specified in dictionaries stored as one attribute of the
   object (e.g., `obj.annotations`).

Finally, the ontology annotations can be defined using namespaces so that the
URIs are shortened (CURIEs). Namespaces are defined for both functions and
data objects using the `namespaces` key in the `__ontology__` dictionary:

* 'namespaces' : dict
   A dictionary where the keys are the names used as prefixes in the
   annotations in `__ontology__` and the values are the prefix URIs to be
   expanded to get the final URI. For example, for an ontology with URIs with
   prefix `http://example.org/ontology#`, such as
   `<http://example.org/ontology#ClassA>` and
   `<http://example.org/ontology#ClassB>` a namespace
   `'ontology'="http://example.org/ontology#"` can be defined such as the
   classes can be referred to as `ontology:ClassA` and `ontology:ClassB`.


Examples
--------

Consider the Python function `process` that takes an input `data`, is
controlled by a parameter `param`, and returns a tuple of two elements:

>>> def process(data, param):
>>>     ...  # process `data` into `output1` and `output2` using `param`
>>>     return output1, output2

To use an ontology defined by the base URI `<http://my-ontology#>` to annotate
the `process` Python function, whose concept is defined by
`<http://my-ontology#ProcessClass>`, and where `data` is represented by
`<http://my-ontology#DataClass>`, `param` by
`<http://my-ontology#ParameterClass>` and the second return element (`output2`)
needs to be annotated with class `<http://my-ontology#ProcessOutputClass>`,
the following dictionary could be inserted into the `process` function object:

>>> process.__ontology__ = {
>>>     'function': "my_ontology:ProcessClass",
>>>     'arguments': {'data': "my_ontology:DataClass",
>>>                   'param': "my_ontology:ParameterClass"},
>>>     'returns': {1: "my_ontology:ProcessOutputClass"},
>>>     'namespaces': {"my_ontology": "http://my-ontology#"}
>>> }

For a Python function `process_container`, represented by
`<http://my-ontology#ProcessContainerClass>` that takes the same inputs but
returns a list of objects that need to be annotated with the ontology class
`<http://my-ontology#ProcessOutputElementClass>`, the following dictionary
could be used:

>>> def process_container(data, param):
>>>     ...  # process `data` into several outputs grouped as a list `output`
>>> return output    # [obj1, obj2, ...]

>>> process_container.__ontology__ = {
>>>     'function': "my_ontology:ProcessContainerClass",
>>>     'arguments': {'data': "my_ontology:DataClass",
>>>                   'param': "my_ontology:ParameterClass"},
>>>     'returns': {'**': "my_ontology:ProcessOutputElementClass"},
>>>     'namespaces': {"my_ontology": "http://my-ontology#"}
>>> }

"""

import rdflib
from copy import deepcopy


# Two types of Python objects can be annotated: functions or data objects.
# For each, specific additional information can also be annotated (e.g.,
# the parameters of a function). This dictionary defines which can be defined
# for each entity, and the strings that are used as keys in the `__ontology__`
# dictionary.
VALID_INFORMATION = {
    'data_object': {'namespaces', 'attributes', 'annotations'},
    'function': {'namespaces', 'arguments', 'returns'}
}
VALID_OBJECTS = set(VALID_INFORMATION.keys())


# Global dictionary to store ontology information during the capture.
# This is used later for the serialization.
ONTOLOGY_INFORMATION = {}


class _OntologyInformation(object):
    """
    Class used to parse information from the `__ontology__` annotation
    dictionary from Python functions or data objects.

    This class provides easy access to the definitions when serializing the
    provenance information with extended ontology annotations. It also manages
    namespaces across different objects and functions, such that no ambiguities
    or multiple definitions are introduced, and the full URIs can be retrieved.

    This class is used internally by Alpaca when serializing the provenance
    as RDF.

    Parameters
    ----------
    obj : object
        Python function or data object with an attribute named `__ontology__`
        that stores a dictionary with specific ontology annotations.
    """

    namespaces = {}

    @classmethod
    def add_namespace(cls, name, uri):
        if name in cls.namespaces:
            if cls.namespaces[name] != uri:
                raise ValueError("Attempting to redefine an existing "
                                 "namespace. This is not allowed as other "
                                 "terms expect a different URI.")
        else:
            cls.namespaces[name] = rdflib.Namespace(uri)

    @classmethod
    def bind_namespaces(cls, namespace_manager):
        for name, namespace in cls.namespaces.items():
            namespace_manager.bind(name, namespace)

    @staticmethod
    def get_ontology_information(obj):
        if hasattr(obj, "__ontology__"):
            return getattr(obj, "__ontology__")
        elif (hasattr(obj, "__wrapped__") and
              hasattr(obj.__wrapped__, "__ontology__")):
            return getattr(obj.__wrapped__, "__ontology__")
        return None

    def __init__(self, obj):

        ontology_info = self.get_ontology_information(obj)
        if ontology_info:
            # An ontology annotation with semantic information is present
            # Store each element inside this object

            for information_type, information in ontology_info.items():
                if information_type in VALID_OBJECTS:
                    # Function or data object URI
                    setattr(self, information_type, information)
                elif information_type == "namespaces":
                    # Add all namespaces, checking for inconsistencies
                    for prefix, uri in information.items():
                        self.add_namespace(prefix, uri)
                else:
                    # Add additional information on the function or data
                    # object
                    setattr(self, information_type, deepcopy(information))

    def has_information(self, information_type):
        return hasattr(self, information_type)

    def get_container_returns(self):
        returns = getattr(self, 'returns', None)
        if returns:
            return [key for key in returns.keys() if isinstance(key, str) and
                    key == '*' * len(key)]
        return None

    def get_uri(self, information_type, element=None):
        if information_type in VALID_OBJECTS:
            # Information on 'function' and 'data_object' are strings or
            # lists, stored directly as attributes
            information_value = getattr(self, information_type)
        else:
            # Specific information of 'function' and 'data_object' are
            # stored in dictionaries (e.g., 'attributes', 'parameters'...)
            information = getattr(self, information_type, None)

            if information is None:
                # No information available
                return None

            # If annotating all elements (e.g., multiple returns in a
            # container). The actual element will not be present, but
            # there will be an entry identified by '*'.
            information_value = information.get(element, None)
            if not information_value:
                return None

        if not isinstance(information_value, list):
            information_value = [information_value]

        # Process URI(s) to get `rdflib.URIRef` elements, resolving any
        # namespace.
        uris = []
        for uri in information_value:
            if (uri[0], uri[-1]) == ("<", ">"):
                # This is a full URI
                uris.append(rdflib.URIRef(uri[1:-1]))
            else:
                # If not full URIs, information must be CURIEs.
                # Get the `URIRef` from the namespace.
                prefix, value = uri.split(":")
                uris.append(self.namespaces[prefix][value])

        if len(uris) == 1:
            # Return annotation with a single URI directly
            return uris[0]
        return uris

    def __repr__(self):
        repr_str = "OntologyInformation("
        information = []
        for obj_type in VALID_OBJECTS:
            if self.has_information(obj_type):
                information.append(f"{obj_type}='{getattr(self, obj_type)}'")
                for specific_information in \
                        sorted(VALID_INFORMATION[obj_type]):
                    if self.has_information(specific_information):
                        specific_info = getattr(self, specific_information)
                        info_str = str(specific_info) \
                            if not isinstance(specific_info, str) else \
                            f"'{specific_info}'"
                        information.append(
                            f"{specific_information}={info_str}")
                repr_str = f"{repr_str}{', '.join(information)})"
        return repr_str



"""
This module implements a function decorator to support provenance capture
during the execution of analysis scripts in Python.
"""

from functools import wraps
import itertools
from collections.abc import Iterable
from collections import defaultdict

from importlib.metadata import version, PackageNotFoundError
import inspect
import ast
import datetime
import logging
import uuid

from alpaca.alpaca_types import FunctionExecution, FunctionInfo, Container
from alpaca.data_information import _ObjectInformation, _FileInformation
from alpaca.code_analysis.ast import _CallAST
from alpaca.code_analysis.source_code import _SourceCode
from alpaca.serialization import AlpacaProvDocument
from alpaca.serialization.identifiers import _get_function_name
from alpaca.utils.files import RDF_FILE_FORMAT_MAP
from alpaca.settings import _ALPACA_SETTINGS
from alpaca.ontology.annotation import _OntologyInformation, ONTOLOGY_INFORMATION

from pprint import pprint


VAR_POSITIONAL = inspect.Parameter.VAR_POSITIONAL
COMPREHENSION_FRAMES = ("<listcomp>", "<dictcomp>", "<setcomp>")


# Create logger and set configuration
logger = logging.getLogger(__file__)
log_handler = logging.StreamHandler()
log_handler.setFormatter(logging.Formatter("[%(asctime)s] alpaca.decorator -"
                                           " %(levelname)s: %(message)s"))
logger.addHandler(log_handler)
logger.propagate = False


class Provenance(object):
    """
    Class to capture and store provenance information in Python scripts.

    The class is a callable object, to be used as a decorator to every function
    call from the script that will be tracked.

    Parameters
    ----------
    inputs : list of str
        Names of the arguments that are considered inputs to the function.
        An input is a variable or value with which the function will perform
        some computation or action. Arguments that only control the behavior
        of the function are considered parameters. The names can be for both
        positional or keyword arguments. Every argument that is not named in
        `inputs`, `container_input`, `file_input` or `file_output` will be
        considered as a parameter. If None, this parameter is ignored. If a
        function does not take any input (e.g., functions that generate data),
        `inputs` can be set to an empty list or None.
    file_input : list of str, optional
        Names of the arguments that represent file(s) read from the disk by
        the function. Their hashes will be computed and stored.
        Default: None
    file_output : list of str, optional
        Names of the arguments that represent file(s) write to the disk by
        the function. The hashes will be computed and stored.
        Default: None
    container_input : list of str, optional
        Names of the arguments that are containers of data (e.g., a list with
        data structures used by the function). Alpaca will track and identify
        the elements inside the container, instead of the container itself.
        Default: None
    container_output : bool or int or tuple, optional
        The function outputs data inside a container (e.g., a list).

        If True, Alpaca will track and identify the elements inside the
        container, instead of the container itself. It will iterate over the
        function output object and identify the individual elements. However,
        for dictionary outputs, the dictionary object is identified together
        with its elements, to retain information on the keys. For other
        containers, the container object is not identified.

        If an integer, this defines a multiple-level (nested) container. The
        number defines the depth for which to identify and serialize the
        objects. In this case, the function output object will always be
        identified together with the element tree. For instance, consider
        the two-level list `L = [[obj1, obj2], [obj3, obj4]]`. With
        `container_output=0`, there will be a single function output node for
        list `L`. Starting from `L`, there will be two additional nodes for
        each of the inner lists (`L[0]` and `L[1]`, i.e., all elements from
        level zero). With `container_output=1`, there will be a single
        function output node for list `L`. Starting from `L`, there
        will be two additional nodes for each of the inner lists (`L[0]` and
        `L[1]`). Finally, starting from each inner list, there will be output
        nodes for `obj1` and `obj2` (linked to `L[0]`) and for `obj3` and
        `obj4` (linked to `L[1]`). Therefore, all elements from level one are
        identified, and linked to the respective elements from level zero.

        If a tuple, this defines a range of the levels in a nested container
        to consider when identifying the objects output by the function. For
        example, taking the same list above, a `container_output=(0, 1)` will
        start from level zero and stop at the elements from level
        one (similar to `container_output=1`). With `container_output=(1, 1)`,
        the first level will be ignored as function output. The function will
        have two output nodes (directly for `L[0]` and `L[1]`). Starting from
        each inner list, there will be output nodes for `obj1` and `obj2`
        (linked to `L[0]`) and for `obj3` and `obj4` (linked to `L[1]`).
        Therefore, the first level (zero) of the container is ignored, and only
        elements from level one are described. The range feature is useful for
        functions where the relevant outputs are containers whose elements
        should also be described, but those containers are grouped inside a
        single return list instead of the function returning a tuple with the
        containers.

        It is important to note that all levels identified as integers or
        range tuples should point to levels in the nested-container that
        contain iterables. For example, in the list `L` above, the level 2
        are the objects `objX`. If `container_output=2`, Alpaca will try to
        iterate over each `objX` and describe their elements. If they are
        not iterable, an error will be raised.

        Default: False

    Attributes
    ----------
    active : bool
        If True, provenance tracking is active.
        If False, provenance tracking is suspended.
        This attribute is set using the :func:`activate`/:func:`deactivate`
        interface functions.
    history : list of FunctionExecution
        All events that were tracked. Each function call is structured in a
        named tuple `FunctionExecution` that stores:

        * 'function': `FunctionInfo` named tuple;
        * 'inputs': `dict` with the `DataObject` or `File` named tuples
          associated with every input value to the function;
        * 'params': `dict` with the positional/keyword argument names that are
          not data/file input/file output as keys. Values are the value of each
          argument as passed to the function call;
        * 'output': `dict` with the `DataObject` or `File` named tuples
          associated with the values returned by the function or files written
          to the disk;
        * 'arg_map': names of the positional arguments;
        * 'kwarg_map': names of the keyword arguments;
        * 'call_ast': `ast.AST` object containing the Abstract Syntax Tree
          of the code that generated the function call.
        * 'code_statement': `str` with the code statement calling the function.
        * 'time_stamp_start', 'time_stamp_end': `str` with the ISO
          representation of the start and end times of the function execution;
        * 'return_targets': names of the variables that store the function
          output(s) in the source code;
        * 'order': integer defining the order of this function call in the
           whole tracking history.
        * 'execution_id': `str` with the UUID of the particular function
           execution tracked.
    source_file : str
        Path to the script file being tracked.
    session_id : str
        Unique identifier (UUID) for this script execution.
    inputs : list
        Names of the function arguments that are considered inputs.
    file_inputs : list
        Names of the function arguments that are considered file inputs.
    file_outputs : list
        Names of the function arguments that are considered file outputs.
    container_inputs : list
        Names of the function arguments that are considered containers of
        data.
    container_output : bool
        True if the function outputs data in a container.

    Raises
    ------
    ValueError
        If `inputs` is not a list or not None.

    """

    active = False
    history = []
    session_id = None
    script_info = None

    source_file = None
    calling_frame = None

    _call_count = 0

    def __init__(self, inputs, file_input=None, file_output=None,
                 container_input=None, container_output=False):
        if inputs is None:
            inputs = []
        if file_input is None:
            file_input = []
        if file_output is None:
            file_output = []
        if container_input is None:
            container_input = []

        if not isinstance(inputs, list):
            raise ValueError("`inputs` must be a list")

        # Store the names of the arguments that are file input/outputs
        # or container inputs
        self.file_inputs = [f_input for f_input in file_input
                            if f_input is not None]
        self.file_outputs = [f_output for f_output in file_output
                             if f_output is not None]
        self.container_inputs = [c_input for c_input in container_input
                                 if c_input is not None]

        # Store the names of arguments that are inputs
        self.inputs = inputs

        self.container_output = False
        self._tracking_container_output = False
        if isinstance(container_output, bool):
            self._tracking_container_output = container_output
            self.container_output = container_output
        elif isinstance(container_output, tuple):
            self._tracking_container_output = len(container_output) == 2
            self.container_output = container_output
        elif isinstance(container_output, int):
            self._tracking_container_output = container_output >= 0
            self.container_output = (0, container_output)

    def _insert_static_information(self, tree, data_info, function,
                                   time_stamp):
        # Use an `ast.NodeVisitor` to find the `Call` node that corresponds to
        # the current `FunctionExecution`. It will fetch static relationships
        # between variables and attributes, and link to the inputs and outputs
        # of the function. The `data_info` object is passed, to use hash
        # memoization in case the hash of some object is already computed for
        # this call.
        ast_visitor = _CallAST(provenance_tracker=self, data_info=data_info,
                               function=function, time_stamp=time_stamp)
        ast_visitor.visit(tree)

    @staticmethod
    def _process_input_arguments(function, args, kwargs):
        # Inspect the arguments to extract the ones defined as inputs.
        # Values are stored in a dictionary with the argument name as key.
        # If signature inspection is not possible, the inputs are stored by
        # order in the function call, with the index as keys. The function
        # also returns the parameters (arguments that are not inputs), with
        # their default values.

        # Initialize dictionaries and lists
        input_data = {}
        input_args_names = []
        input_kwargs_names = []

        try:
            # Get the function signature and bind the arguments, obtaining a
            # dictionary with argument name as keys and argument value as
            # values
            fn_sig = inspect.signature(function)
            func_parameters = fn_sig.bind(*args, **kwargs)

            # Get the default argument values, to store them in case they
            # were not passed in the call
            default_args = {k: v.default
                            for k, v in fn_sig.parameters.items()
                            if v.default is not inspect.Parameter.empty}

            # For each item in the bound arguments dictionary...
            for arg_name, arg_value in func_parameters.arguments.items():

                # Get the description of the current argument by its name
                cur_parameter = \
                    func_parameters.signature.parameters[arg_name]

                # If this argument is one of possible default values, remove
                # it, since the user has passed a value explicitly
                if arg_name in default_args:
                    default_args.pop(arg_name)

                # If the argument is variable positional (i.e., *arg) we will
                # store its value in the input dictionary as the Container
                # named tuple. This signals that this argument's value is
                # multiple. Otherwise, we just store the argument value.
                if cur_parameter.kind != VAR_POSITIONAL:
                    input_data[arg_name] = arg_value
                else:
                    # Variable positional arguments are stored as
                    # the named tuple Container.
                    input_data[arg_name] = Container(arg_value)

                # Store the argument name in the appropriate list
                if arg_name in kwargs:
                    input_kwargs_names.append(arg_name)
                else:
                    input_args_names.append(arg_name)

            # Add the default argument names to the list of kwargs names
            input_kwargs_names.extend(default_args.keys())

        except ValueError:
            # Can't inspect signature. Append args/kwargs by order
            for arg_index, arg in enumerate(args):
                input_data[arg_index] = arg
                input_args_names.append(arg_index)

            # Keyword arguments index start after the last positional argument
            kwarg_start = len(input_data)
            for kwarg_index, kwarg in enumerate(kwargs,
                                                start=kwarg_start):
                input_data[kwarg_index] = kwarg
                input_kwargs_names.append(kwarg_index)

            # No default arguments
            default_args = {}

        return input_data, input_args_names, input_kwargs_names, default_args

    @staticmethod
    def _get_module_version(module):

        if not (module is None or module.startswith("__main__")):
            # User-defined functions in the running script do not have a
            # version
            package = module.split(".")[0]
            try:
                return version(package)
            except PackageNotFoundError:
                # When running unit tests or using user-defined functions
                # imported from a source file
                return ""
        return ""

    def _get_calling_line_number(self, frame):
        # Get the line number of the current call.
        # For that, we need to find the frame containing the call, starting
        # from `frame`, which is the current frame being executed.
        lineno = None

        # Extract information and calling function name in `frame`
        frame_info = inspect.getframeinfo(frame)
        function_name = frame_info.function

        if function_name in COMPREHENSION_FRAMES:
            # For comprehensions, we need to check the frame above,
            # as this creates a function named <*comp>. We use a while loop
            # in case of nested comprehensions.
            while function_name in COMPREHENSION_FRAMES:
                frame = frame.f_back
                frame_info = inspect.getframeinfo(frame)
                function_name = frame_info.function
        elif function_name == 'wrapper':
            # For functions with a decorator, we need to skip the decorator
            frame = frame.f_back
            frame_info = inspect.getframeinfo(frame)
            function_name = frame_info.function

        # If the frame corresponds to the script file and the tracked function,
        # we get the line number
        if (frame_info.filename == self.source_file and
                function_name == self._source_code.source_name):
            lineno = frame.f_lineno

        return lineno

    @staticmethod
    def _is_class_constructor(function_name):
        names = function_name.split(".")
        return len(names) == 2 and names[-1] == "__init__"

    @staticmethod
    def _is_static_method(function, function_name):
        if type(function).__qualname__ == "method_descriptor":
            # Ignore method descriptors
            return False
        name = function_name.rsplit('.', 1)[-1]
        cls = inspect._findclass(function)
        if cls is not None:
            method = inspect.getattr_static(cls, name)
            return isinstance(method, staticmethod)
        return False

    def _capture_code_and_function_provenance(self, lineno, function):

        # 1. Capture Abstract Syntax Tree (AST) of the call to the
        # function. We need to check the source code in case the
        # call spans multiple lines. In this case, we fetch the
        # full statement using the code analyzer.
        source_line = \
            self._source_code.extract_multiline_statement(lineno)
        ast_tree = ast.parse(source_line)
        logger.debug(f"Line {lineno} -> {source_line}")

        # 2. Check if there is an assignment to one or more
        # variables. This will be used to identify if there are
        # multiple output nodes. This is needed because just
        # checking if `function_output` is tuple does not work if
        # the function is actually returning a tuple.
        return_targets = []
        if isinstance(ast_tree.body[0], ast.Assign):
            assign_target = ast_tree.body[0].targets[0]
            if isinstance(assign_target, ast.Tuple):
                return_targets = [target.id for target in
                                  assign_target.elts]
            elif isinstance(assign_target, ast.Name):
                return_targets = [assign_target.id]
            else:
                # This branch should not be reachable
                raise ValueError("Unknown assign target!")

        # 3. Extract function name and information
        function_name = function.__qualname__
        module = None
        try:
            module = getattr(function, '__module__')
        except AttributeError:
            # Case of method descriptors
            if type(function).__qualname__ == "method_descriptor":
                module = getattr(function.__objclass__, '__module__')

        module_version = self._get_module_version(module)
        function_info = FunctionInfo(name=function_name, module=module,
                                     version=module_version)

        function_id = _get_function_name(function_info)
        if not ONTOLOGY_INFORMATION.get(function_id):
            if _OntologyInformation.get_ontology_information(function):
                ONTOLOGY_INFORMATION[function_id] = \
                    _OntologyInformation(function)

        return source_line, ast_tree, return_targets, function_info

    def _capture_input_and_parameters_provenance(self, function, args, kwargs,
        ast_tree, function_info, time_stamp_start, builtin_object_hash,
        store_values):

        # 1. Extract the parameters passed to the function and store them in
        # the `input_data` dictionary.
        # Two separate lists with the names according to the arg/kwarg order
        # are also constructed, to map to the `args` and `keywords` fields
        # of the AST nodes. Also, the list of all arguments whose values taken
        # are defaults is returned as the `default_args` dictionary.

        input_data, input_args_names, input_kwargs_names, default_args = \
            self._process_input_arguments(function, args, kwargs)

        # 2. Create parameters/input descriptions for the graph.
        # Here the inputs, but not the parameters passed to the function, are
        # hashed using the `_ObjectInformation` object.
        # Inputs are defined by the parameter `inputs` when initializing the
        # decorator, and stored as the attribute `inputs`. If one parameter
        # is defined as a `file_input` in the initialization, a hash to the
        # file is obtained using the `_FileInformation` object. If one
        # parameter is defined as `container_input` in the initialization, its
        # elements are hashed and stored if the value is iterable.
        # After this step, all hashes and metadata of input parameters/files
        # are going to be stored in the dictionary `inputs`.

        data_info = _ObjectInformation(use_builtin_hash=builtin_object_hash,
                                       store_values=store_values)

        # Initialize parameter list with all default arguments that were not
        # passed to the function
        parameters = default_args

        inputs = {}
        for key, input_value in input_data.items():
            if key in self.inputs:
                if isinstance(input_value, Container):
                    # If the argument is multiple, hash each value
                    # tuple and store them inside a `Container` namedtuple so
                    # that we know this is a multiple input
                    var_input_list = []
                    for var_arg in input_value.elements:
                        var_input_list.append(data_info.info(var_arg))
                    inputs[key] = Container(tuple(var_input_list))
                else:
                    inputs[key] = data_info.info(input_value)

            elif key in self.file_inputs:
                # Input is from a file. Hash using `_FileInformation`
                inputs[key] = _FileInformation(input_value).info()

            elif key in self.container_inputs and \
                    (isinstance(input_value, Iterable) or
                     hasattr(input_value, "__getitem__")):
                # This is a container. Iterate over elements and store inside
                # a `Container` namedtuple
                container_elements = [data_info.info(element)
                                      for element in input_value]
                inputs[key] = Container(tuple(container_elements))

            elif key not in self.file_outputs:
                # The remainder argument is also not an output file, so this
                # is a parameter to the function.
                parameters[key] = input_value

        # 3. Analyze AST and fetch static relationships in the
        # input/output and other variables/objects in the script
        self._insert_static_information(tree=ast_tree, data_info=data_info,
                                        function=function_info.name,
                                        time_stamp=time_stamp_start)

        return inputs, parameters, input_args_names, input_kwargs_names, \
            input_data

    def _add_container_relationships(self, container, data_info, level,
                                     time_stamp_start, execution_id):
        # For every element of the container, add a subscript relationship
        # This ensures that the indexing information is captured and
        # described. The hash memoization will prevent multiple hashing.
        input_object = data_info.info(container)

        if isinstance(container, dict):
            iterator = container.items()
        elif isinstance(container, Iterable) or \
                hasattr(container, "__getitem__"):
            iterator = enumerate(container)
        else:
            iterator = enumerate([container])

        for index, element in iterator:
            output_object = data_info.info(element)
            self.history.append(
                FunctionExecution(
                    function=FunctionInfo(name='subscript', module="",
                                          version=""),
                    input={0: input_object},
                    params={'index': index},
                    output={0: output_object},
                    arg_map=None,
                    kwarg_map=None,
                    call_ast=None,
                    code_statement=None,
                    time_stamp_start=time_stamp_start,
                    time_stamp_end=time_stamp_start,
                    return_targets=[],
                    order=None,
                    execution_id=execution_id))

            # If multilevel requested, process the next level.
            # This will work whether the main container is a dictionary or
            # other iterable.
            if (level is not None and
                    level < max(self.container_output) and
                    (isinstance(element, Iterable) or
                     hasattr(container, "__getitem__"))):
                self._add_container_relationships(element, data_info,
                                                  level + 1,
                                                  time_stamp_start,
                                                  execution_id)

        return input_object

    def _capture_container_output(self, function_output, data_info,
                                  time_stamp_start, execution_id):
        level = None if isinstance(self.container_output, bool) else 0

        if isinstance(function_output, dict):
            container_info = self._add_container_relationships(
                function_output, data_info, level, time_stamp_start,
                execution_id)
            return {0: container_info}
        elif level is not None:
            if not self.container_output or min(self.container_output) == 0:
                # Starting from zero
                container_info = self._add_container_relationships(
                    function_output, data_info, level, time_stamp_start,
                    execution_id)
                return {0: container_info}
            else:
                # Process range starting from other level
                elements = function_output
                start_level = min(self.container_output) - 1
                for level in range(start_level):
                    # Unpack all elements until the requested start level
                    elements = itertools.chain(*elements)
                return {idx: self._add_container_relationships(
                    element, data_info, start_level + 1, time_stamp_start,
                    execution_id) for idx, element in enumerate(elements)}

        # Process simple container.
        # The container object will not be identified.
        return {index: data_info.info(item)
                for index, item in enumerate(function_output)}

    def _capture_output_provenance(self, function_output, return_targets,
                                   input_data, builtin_object_hash,
                                   time_stamp_start, execution_id,
                                   store_values, constructed_object=None):

        # In case in-place operations were performed, lets not use
        # memoization
        data_info = _ObjectInformation(use_builtin_hash=builtin_object_hash,
                                       store_values=store_values)

        # 6. Create hash for the output using `_ObjectInformation` to follow
        # individual returns. The hashes will be stored in the `outputs`
        # dictionary, with the index as the order of each returned object.
        # If the decorator was initialized with `container_output=True`, the
        # elements of the output will be hashed, if iterable.

        # If this was the `__init__` method, we do not consider the
        # None object returned, as the call is returning the
        # constructed object instance
        function_output = function_output \
            if constructed_object is None else constructed_object

        if self._tracking_container_output and \
                (isinstance(function_output, Iterable) or
                        hasattr(function_output, "__getitem__")):
            outputs = self._capture_container_output(function_output,
                                                     data_info,
                                                     time_stamp_start,
                                                     execution_id)
        else:
            if len(return_targets) < 2:
                function_output = [function_output]

            outputs = {index: data_info.info(item)
                       for index, item in enumerate(function_output)}

        # If there is a file output as defined in the decorator
        # initialization, create the hash and add as output using
        # `_FileInformation`. These outputs will be identified by the key
        # `file.X`, where X is an integer with the order of the file output
        if self.file_outputs:
            for idx, file_output in enumerate(self.file_outputs):
                outputs[f"file.{idx}"] = \
                    _FileInformation(input_data[file_output]).info()

        return outputs

    def __call__(self, function):

        @wraps(function)
        def wrapped(*args, **kwargs):

            builtin_object_hash = _ALPACA_SETTINGS[
                'use_builtin_hash_for_module']
            store_values = _ALPACA_SETTINGS['store_values']
            logging.debug(f"Builtin object hash: {builtin_object_hash}")

            lineno = None

            # If capturing provenance, get the code, function, inputs and
            # parameter information, before executing the function
            if Provenance.active:

                # For functions that are used inside other decorated functions,
                # or recursively, check if the calling frame is the one being
                # tracked. If this call comes from the frame tracked, we will
                # get the line number. Otherwise, the line number will be
                # None, and the provenance tracking block will be skipped.
                try:
                    frame = inspect.currentframe().f_back
                    lineno = self._get_calling_line_number(frame)
                finally:
                    del frame

                if lineno:
                    # Get the start time stamp
                    time_stamp_start = datetime.datetime.utcnow().isoformat()

                    # Increment the global call counter
                    Provenance._call_count += 1

                    # Create execution ID
                    execution_id = str(uuid.uuid4())

                    # Capture code and function information
                    source_line, ast_tree, return_targets, function_info = \
                        self._capture_code_and_function_provenance(
                            lineno=lineno, function=function)

                    # Capture input and parameter information
                    inputs, parameters, input_args_names, \
                        input_kwargs_names, input_data = \
                            self._capture_input_and_parameters_provenance(
                                function=function, args=args, kwargs=kwargs,
                                ast_tree=ast_tree, function_info=function_info,
                                time_stamp_start=time_stamp_start,
                                builtin_object_hash=builtin_object_hash,
                                store_values=store_values)

            # Call the function
            function_output = function(*args, **kwargs)

            # If capturing provenance, resume capturing the output information
            if Provenance.active and lineno:

                # If this was a constructor, we work with the object defined
                # by the `self` argument
                constructed_object = None
                if self._is_class_constructor(function_info.name):
                    # Capture information of the `self` object
                    constructed_object = input_data.get('self', None)

                    # Remove `self` from the parameters, in case it was not
                    # specified as an input
                    if 'self' in parameters:
                        parameters.pop('self')

                # Capture output information
                outputs = self._capture_output_provenance(
                    function_output=function_output,
                    return_targets=return_targets, input_data=input_data,
                    builtin_object_hash=builtin_object_hash,
                    time_stamp_start=time_stamp_start,
                    execution_id=execution_id,
                    store_values=store_values,
                    constructed_object=constructed_object)

                # Get the end time stamp
                time_stamp_end = datetime.datetime.utcnow().isoformat()

                # Create FunctionExecution tuple
                function_execution = FunctionExecution(
                    function=function_info,
                    input=inputs,
                    params=parameters,
                    output=outputs,
                    arg_map=input_args_names,
                    kwarg_map=input_kwargs_names,
                    call_ast=ast_tree,
                    code_statement=source_line,
                    time_stamp_start=time_stamp_start,
                    time_stamp_end=time_stamp_end,
                    return_targets=return_targets,
                    order=Provenance._call_count,
                    execution_id=execution_id)

                # Add to the history.
                # The history will be the base to generate the graph and
                # PROV document.
                Provenance.history.append(function_execution)

            return function_output

        # If the function is decorated with `staticmethod`, restore the
        # decorator (otherwise `self` will be passed as first argument when
        # calling the function)
        if self._is_static_method(function, function.__qualname__):
            return staticmethod(wrapped)

        return wrapped

    @classmethod
    def _get_script_variable(cls, name):
        # Access to variable values in the tracked code by name.
        return cls.calling_frame.f_locals[name]

    @classmethod
    def _set_calling_frame(cls, frame):
        """
        This method stores the frame of the code being tracked, and
        extract several information that is needed for capturing provenance.

        A `_SourceCode` object is created, to provide an interface to
        retrieve information from the code (e.g., statements given a line
        number).

        It also initializes a unique ID for the script run, and stores the
        information regarding the script file (`File` named tuple).

        Parameters
        ----------
        frame : inspect.frame
            Frame object returned by the `inspect` module. This must
            correspond to the namespace where provenance tracking was
            activated. This is automatically fetched by the interface function
            :func:`activate`.
        """

        # Store the reference to the calling frame
        cls.calling_frame = frame

        # Get the file name and function associated with the frame
        cls.source_file = inspect.getfile(frame)

        # Create a _SourceCode instance with the frame information,
        # so that we can capture provenance information later
        cls._source_code = _SourceCode(frame)

        # Create a unique identifier for the session and store script info
        cls.session_id = str(uuid.uuid4())
        cls.script_info = _FileInformation(cls.source_file).info()

    @classmethod
    def get_prov_info(cls, show_progress=False):
        """
        Returns the RDF representation of the captured provenance information
        according to the Alpaca ontology based on the W3C PROV-O.

        Parameters
        ----------
        show_progress : bool, optional
            If True, show a bar with the progress of the conversion of the
            captured provenance information to RDF.
            Default: False

        Returns
        -------
        serialization.AlpacaProvDocument
        """

        prov_document = AlpacaProvDocument()
        prov_document.add_history(script_info=cls.script_info,
                                  session_id=cls.session_id,
                                  history=cls.history,
                                  show_progress=show_progress)
        return prov_document

    @classmethod
    def clear(cls):
        """
        Clears all the history, resets the execution counter to zero,
        and removes script information.
        """
        cls.history.clear()
        cls._call_count = 0
        cls.script_info = None


##############################################################################
# Interface functions
##############################################################################

def activate(clear=False):
    """
    Activates provenance tracking within the script.

    Parameters
    ----------
    clear : bool, optional
        If True, the history is cleared and execution counter is reset to
        zero.
        Default: False
    """
    if clear:
        Provenance.clear()

    # To access variables in the same namespace where the function is called,
    # and get information from the source code, the previous frame in the
    # stack needs to be saved.
    Provenance._set_calling_frame(inspect.currentframe().f_back)
    Provenance.active = True


def deactivate():
    """
    Deactivates provenance tracking within Elephant.
    """
    Provenance.calling_frame = None
    Provenance.active = False


def print_history():
    """
    Print all executions in the provenance track history.
    """
    pprint(Provenance.history)


def save_provenance(file_name=None, file_format='ttl',show_progress=False):
    """
    Serialize provenance information to RDF according to the Alpaca ontology
    based on the W3C PROV Ontology (PROV-O).

    Parameters
    ----------
    file_name : str or Path-like, optional
        Destination file to serialize the provenance information.
        If None, the function will return a string containing the provenance
        information in the specified format.
        Default: None
    file_format : {'json-ld', 'n3', 'nt', 'hext', 'pretty-xml', 'trig', 'turtle', 'longturtle', 'xml', 'ttl', 'rdf', 'json'}
        Format into which the provenance data is serialized. The formats are
        the ones accepted by RDFLib. Some shortucts are defined for common
        file extensions:

        * 'ttl': Turtle
        * 'rdf': RDF/XML
        * 'json': JSON-LD

        Default: 'ttl'
    show_progress : bool, optional
        If True, show a bar with the progress of the serialization to RDF.
        Default: False

    Returns
    -------
    str or None
        If `file_name` is None, the function returns the PROV information as
        a string. If a file destination was informed, the return is None.
    """
    # If provenance was not captured, there will be no information for the
    # script. No information will be serialized.
    if not Provenance.script_info:
        return

    if file_format in RDF_FILE_FORMAT_MAP:
        file_format = RDF_FILE_FORMAT_MAP[file_format]

    prov_document = Provenance.get_prov_info(show_progress=show_progress)
    prov_data = prov_document.serialize(file_name, file_format=file_format)
    return prov_data



"""
Alpaca is a Python package for the capture of provenance information during
the execution of Python scripts that process data.
"""

from .decorator import (Provenance, activate, deactivate, save_provenance,
                        print_history)
from .serialization import AlpacaProvDocument
from .graph import ProvenanceGraph
from .settings import alpaca_setting
from .utils import files



import unittest

import inspect
from functools import wraps
from alpaca.code_analysis.source_code import _SourceCode


# Dummy functions and variables
# They are used inside the tests to provide an executable code that can be
# analyzed by the _SourceCode class that is being tested here

def function_call(*args):
    return


def decorator(function):
    @wraps(function)
    def wrapper(*args, **kwargs):
        return function(*args, **kwargs)
    return wrapper


arg11 = arg12 = arg21 = arg22 = arg31 = arg32 = arg41 = arg42 = arg51 \
    = arg52 = None


# To hold the source code object (dummy for the Provenance decorator)

class Class:
    source_code = None


# Mimics the behavior of the activate function of Alpaca, to get the frame of
# the scope where it was called

def activate():
    Class.source_code = _SourceCode(inspect.currentframe().f_back)


# Expected results for each statement

RES1 = 'res1 = function_call(arg11, arg12)'
RES2 = """res2 = function_call(arg21,
                                 arg22)"""
RES3 = """res3 = \\
                function_call(arg31, arg32)"""
RES4 = 'res4 = function_call(arg41, arg42)'
RES5 = 'res5 = function_call(arg51, arg52)'
RES6 = 'res6 = function_call(arg11, arg21)'


class CodeAnalyzerTestCase(unittest.TestCase):

    def setUp(self):
        Class.source_code = None

    def test_function_main(self):

        def main():
            activate()

            res1 = function_call(arg11, arg12)

            res2 = function_call(arg21,
                                 arg22)

            res3 = \
                function_call(arg31, arg32)

            res4 = function_call(arg41, arg42)
            res5 = function_call(arg51, arg52)

            # With comment
            res6 = function_call(arg11, arg21)

        main()
        source_code = Class.source_code

        expected_statements = {
            59: None,
            60: "activate()",
            61: None,
            62: RES1,
            63: None,
            64: RES2, 65: RES2,
            66: None,
            67: RES3, 68: RES3,
            69: None,
            70: RES4,
            71: RES5,
            72: None,
            73: None,
            74: RES6, 75: None,
        }

        for line, expected_statement in expected_statements.items():
            with self.subTest(f"line: {line}, "
                              f"expected: {expected_statement}"):
                self.assertEqual(
                    source_code.extract_multiline_statement(line),
                    expected_statement
                )

    def test_function_activate_middle(self):

        def main():
            res1 = function_call(arg11, arg12)

            res2 = function_call(arg21,
                                 arg22)

            res3 = \
                function_call(arg31, arg32)
            activate()

            res4 = function_call(arg41, arg42)
            res5 = function_call(arg51, arg52)

            # With comment
            res6 = function_call(arg11, arg21)

        main()
        source_code = Class.source_code

        expected_statements = {
            106: None,
            107: RES1,
            108: None,
            109: RES2, 110: RES2,
            111: None,
            112: RES3, 113: RES3,
            114: "activate()",
            115: None,
            116: RES4,
            117: RES5,
            118: None,
            119: None,
            120: RES6, 121: None
        }

        for line, expected_statement in expected_statements.items():
            with self.subTest(f"line: {line}, "
                              f"expected: {expected_statement}"):
                self.assertEqual(
                    source_code.extract_multiline_statement(line),
                    expected_statement
                )

    def test_function_decorator(self):

        @decorator
        def main():
            activate()

            res1 = function_call(arg11, arg12)

            res2 = function_call(arg21,
                                 arg22)

            res3 = \
                function_call(arg31, arg32)

            res4 = function_call(arg41, arg42)
            res5 = function_call(arg51, arg52)

            # With comment
            res6 = function_call(arg11, arg21)

        main()
        source_code = Class.source_code

        expected_statements = {
            152: None,
            153: "activate()",
            154: None,
            155: RES1,
            156: None,
            157: RES2, 158: RES2,
            159: None,
            160: RES3, 161: RES3,
            162: None,
            163: RES4,
            164: RES5,
            165: None,
            166: None,
            167: RES6, 168: None
        }

        for line, expected_statement in expected_statements.items():
            with self.subTest(f"line: {line}, "
                              f"expected: {expected_statement}"):
                self.assertEqual(
                    source_code.extract_multiline_statement(line),
                    expected_statement
                )


class CodeAnalyzerWithBlocksTestCase(unittest.TestCase):

    def test_for(self):

        def main():
            activate()

            res1 = function_call(arg11, arg12)

            for arg in range(5):
                res2 = function_call(arg21,
                                 arg22)

                res3 = \
                function_call(arg31, arg32)

                res4 = function_call(arg41, arg42)
            res5 = function_call(arg51, arg52)

            # With comment
            res6 = function_call(arg11, arg21)

        main()
        source_code = Class.source_code

        expected_statements = {
            202: None,
            203: "activate()",
            204: None,
            205: RES1,
            206: None, 207: None,
            208: RES2, 209: RES2,
            210: None,
            211: RES3, 212: RES3,
            213: None,
            214: RES4,
            215: RES5,
            216: None,
            217: None,
            218: RES6, 219: None
        }

        for line, expected_statement in expected_statements.items():
            with self.subTest(f"line: {line}, "
                              f"expected: {expected_statement}"):
                self.assertEqual(
                    source_code.extract_multiline_statement(line),
                    expected_statement
                )

    def test_if(self):

        def main():
            activate()

            res1 = function_call(arg11, arg12)

            if arg21 is None:
                res2 = function_call(arg21,
                                 arg22)

            elif arg31 is None:
                res3 = \
                function_call(arg31, arg32)

            else:
                res4 = function_call(arg41, arg42)
            res5 = function_call(arg51, arg52)

            # With comment
            res6 = function_call(arg11, arg21)

        main()
        source_code = Class.source_code

        expected_statements = {
            250: None,
            251: "activate()",
            252: None,
            253: RES1,
            254: None, 255: None,
            256: RES2, 257: RES2,
            258: None, 259: None,
            260: RES3, 261: RES3,
            262: None, 263: None,
            264: RES4,
            265: RES5,
            266: None,
            267: None,
            268: RES6, 269: None
        }

        for line, expected_statement in expected_statements.items():
            with self.subTest(f"line: {line}, "
                              f"expected: {expected_statement}"):
                self.assertEqual(
                    source_code.extract_multiline_statement(line),
                    expected_statement
                )


if __name__ == "__main__":
    unittest.main()



"""
These unit tests are done using the decorator with proper syntactic calls,
i.e., using static relationships in the inputs of the functions.

The files in the `ast` and `static_relationship_tree` files in `code_analysis`
are covered by the tests in this file.
"""

import unittest

import joblib
import datetime
from functools import partial

import numpy as np

from alpaca import Provenance, activate, deactivate
from alpaca.alpaca_types import FunctionInfo, DataObject
from alpaca.ontology.annotation import _OntologyInformation


# Shortcut for SHA1 hashing using joblib
joblib_hash = partial(joblib.hash, hash_name='sha1')


# Define some data for testing

TEST_ARRAY = np.array([1, 2, 3])
TEST_ARRAY_INFO = DataObject(hash=joblib_hash(TEST_ARRAY),
                             hash_method="joblib_SHA1",
                             type="numpy.ndarray", id=id(TEST_ARRAY),
                             details={'shape': (3,), 'dtype': np.int64},
                             value=None)

ELEMENT_0_INFO = DataObject(hash=joblib_hash(TEST_ARRAY[0]),
                            hash_method="joblib_SHA1", type="numpy.int64",
                            id=id(TEST_ARRAY[0]),
                            details={'shape': (), 'dtype': np.int64},
                            value=1)

ELEMENT_1_INFO = DataObject(hash=joblib_hash(TEST_ARRAY[1]),
                            hash_method="joblib_SHA1", type="numpy.int64",
                            id=id(TEST_ARRAY[1]),
                            details={'shape': (), 'dtype': np.int64},
                            value=2)

ELEMENT_2_INFO = DataObject(hash=joblib_hash(TEST_ARRAY[2]),
                            hash_method="joblib_SHA1", type="numpy.int64",
                            id=id(TEST_ARRAY[2]),
                            details={'shape': (), 'dtype': np.int64},
                            value=3)

TEST_DICT = {'numbers': TEST_ARRAY}
TEST_DICT_INFO = DataObject(hash=joblib_hash(TEST_DICT),
                            hash_method="joblib_SHA1",
                            type="builtins.dict", id=id(TEST_DICT),
                            details={}, value=None)


# To test attributes
class ContainerOfArray:
    def __init__(self, array):
        self.array = array


# To test attribute calls
class ObjectWithMethod:
    def add_numbers(self, array):
        return np.sum(array)
ObjectWithMethod.add_numbers = Provenance(inputs=['self', 'array'])(ObjectWithMethod.add_numbers)

class CustomObject:
    def __init__(self, data):
        self.data = data
CustomObject.__init__ = Provenance(inputs=['data'])(CustomObject.__init__)

# Define some test functions to use different relationships

@Provenance(inputs=['num1', 'num2'])
def add_numbers(num1, num2):
    return num1 + num2


@Provenance(inputs=['array'])
def add_numbers_array(array):
    return np.sum(array)


@Provenance(inputs=['array'])
def get_number(array, index):
    return array[index]


# Function to help verifying FunctionExecution tuples
# For subscripting, the IDs usually change, so we need to ignore them
def _check_function_execution(actual, exp_function, exp_input, exp_params,
                              exp_output, exp_arg_map, exp_kwarg_map,
                              exp_code_stmnt, exp_return_targets, exp_order,
                              test_case):

    data_object_attributes = ('hash', 'hash_method', 'type', 'details',
                              'value')

    # Check function
    test_case.assertTupleEqual(actual.function, exp_function)

    # Check inputs
    for input_arg, value in actual.input.items():
        test_case.assertTrue(input_arg in exp_input)
        for attr in data_object_attributes:
            test_case.assertEqual(getattr(value, attr),
                                  getattr(exp_input[input_arg], attr))

    # Check parameters
    test_case.assertDictEqual(actual.params, exp_params)

    # Check outputs
    for output, value in actual.output.items():
        test_case.assertTrue(output in exp_output)
        for attr in data_object_attributes:
            test_case.assertEqual(getattr(value, attr),
                                  getattr(exp_output[output], attr))

    # Check args and kwargs
    if actual.arg_map is not None:
        test_case.assertListEqual(actual.arg_map, exp_arg_map)
    else:
        test_case.assertIsNone(exp_arg_map)

    if actual.kwarg_map is not None:
        test_case.assertListEqual(actual.kwarg_map, exp_kwarg_map)
    else:
        test_case.assertIsNone(exp_kwarg_map)

    # Check other information
    test_case.assertEqual(actual.code_statement, exp_code_stmnt)
    test_case.assertListEqual(actual.return_targets, exp_return_targets)
    test_case.assertEqual(actual.order, exp_order)
    test_case.assertNotEqual(actual.execution_id, "")

    # Check time stamps are valid ISO dates
    test_case.assertIsInstance(
        datetime.datetime.fromisoformat(actual.time_stamp_start),
        datetime.datetime)
    test_case.assertIsInstance(
        datetime.datetime.fromisoformat(actual.time_stamp_end),
        datetime.datetime)


class ProvenanceDecoratorStaticRelationshipsTestCase(unittest.TestCase):

    def test_subscript_index(self):
        activate(clear=True)
        source_array = TEST_ARRAY
        res = add_numbers(source_array[0], source_array[1])
        deactivate()

        self.assertEqual(len(Provenance.history), 3)

        expected_output = DataObject(
            hash=joblib_hash(TEST_ARRAY[0]+TEST_ARRAY[1]),
            hash_method="joblib_SHA1",
            type="numpy.int64", id=id(res),
            details={'shape': (), 'dtype': np.int64}, value=3)

        _check_function_execution(
            actual=Provenance.history[0],
            exp_function=FunctionInfo('subscript', '', ''),
            exp_input={0: TEST_ARRAY_INFO},
            exp_params={'index': 0},
            exp_output={0: ELEMENT_0_INFO},
            exp_arg_map=None, exp_kwarg_map=None, exp_code_stmnt=None,
            exp_return_targets=[], exp_order=None,
            test_case=self)

        _check_function_execution(
            actual=Provenance.history[1],
            exp_function=FunctionInfo('subscript', '', ''),
            exp_input={0: TEST_ARRAY_INFO},
            exp_params={'index': 1},
            exp_output={0: ELEMENT_1_INFO},
            exp_arg_map=None, exp_kwarg_map=None, exp_code_stmnt=None,
            exp_return_targets=[], exp_order=None,
            test_case=self)

        _check_function_execution(
            actual=Provenance.history[2],
            exp_function=FunctionInfo('add_numbers',
                                      'test_code_analysis', ''),
            exp_input={'num1': ELEMENT_0_INFO, 'num2': ELEMENT_1_INFO},
            exp_params={},
            exp_output={0: expected_output},
            exp_arg_map=['num1', 'num2'],
            exp_kwarg_map=[],
            exp_code_stmnt="res = add_numbers(source_array[0], source_array[1])",
            exp_return_targets=['res'],
            exp_order=1,
            test_case=self)

    def test_subscript_negative_index(self):
        activate(clear=True)
        source_array = TEST_ARRAY
        res = add_numbers(source_array[-1], source_array[-2])
        deactivate()

        self.assertEqual(len(Provenance.history), 3)

        expected_output = DataObject(
            hash=joblib_hash(TEST_ARRAY[-1]+TEST_ARRAY[-2]),
            hash_method="joblib_SHA1",
            type="numpy.int64", id=id(res),
            details={'shape': (), 'dtype': np.int64}, value=5)

        _check_function_execution(
            actual=Provenance.history[0],
            exp_function=FunctionInfo('subscript', '', ''),
            exp_input={0: TEST_ARRAY_INFO},
            exp_params={'index': -1},
            exp_output={0: ELEMENT_2_INFO},
            exp_arg_map=None, exp_kwarg_map=None, exp_code_stmnt=None,
            exp_return_targets=[], exp_order=None,
            test_case=self)

        _check_function_execution(
            actual=Provenance.history[1],
            exp_function=FunctionInfo('subscript', '', ''),
            exp_input={0: TEST_ARRAY_INFO},
            exp_params={'index': -2},
            exp_output={0: ELEMENT_1_INFO},
            exp_arg_map=None, exp_kwarg_map=None, exp_code_stmnt=None,
            exp_return_targets=[], exp_order=None,
            test_case=self)

        _check_function_execution(
            actual=Provenance.history[2],
            exp_function=FunctionInfo('add_numbers',
                                      'test_code_analysis', ''),
            exp_input={'num1': ELEMENT_2_INFO, 'num2': ELEMENT_1_INFO},
            exp_params={},
            exp_output={0: expected_output},
            exp_arg_map=['num1', 'num2'],
            exp_kwarg_map=[],
            exp_code_stmnt="res = add_numbers(source_array[-1], source_array[-2])",
            exp_return_targets=['res'],
            exp_order=1,
            test_case=self)

    def test_subscript_slice(self):
        activate(clear=True)
        source_array = TEST_ARRAY
        res = add_numbers_array(source_array[0:2])
        deactivate()

        self.assertEqual(len(Provenance.history), 2)

        expected_output = DataObject(
            hash=joblib_hash(TEST_ARRAY[0]+TEST_ARRAY[1]),
            hash_method="joblib_SHA1",
            type="numpy.int64", id=id(res),
            details={'shape': (), 'dtype': np.int64}, value=3)

        expected_slice_output = DataObject(
            hash=joblib_hash(TEST_ARRAY[0:2]), hash_method="joblib_SHA1",
            type="numpy.ndarray", id=id(TEST_ARRAY[0:2]),
            details={'shape': (2,), 'dtype': np.int64}, value=None)

        _check_function_execution(
            actual=Provenance.history[0],
            exp_function=FunctionInfo('subscript', '', ''),
            exp_input={0: TEST_ARRAY_INFO},
            exp_params={'slice': '0:2'},
            exp_output={0: expected_slice_output},
            exp_arg_map=None, exp_kwarg_map=None, exp_code_stmnt=None,
            exp_return_targets=[], exp_order=None,
            test_case=self)

        _check_function_execution(
            actual=Provenance.history[1],
            exp_function=FunctionInfo('add_numbers_array',
                                      'test_code_analysis', ''),
            exp_input={'array': expected_slice_output},
            exp_params={},
            exp_output={0: expected_output},
            exp_arg_map=['array'],
            exp_kwarg_map=[],
            exp_code_stmnt="res = add_numbers_array(source_array[0:2])",
            exp_return_targets=['res'],
            exp_order=1,
            test_case=self)

    def test_subscript_slice_no_start(self):
        activate(clear=True)
        source_array = TEST_ARRAY
        res = add_numbers_array(source_array[:2])
        deactivate()

        self.assertEqual(len(Provenance.history), 2)

        expected_output = DataObject(
            hash=joblib_hash(TEST_ARRAY[0]+TEST_ARRAY[1]),
            hash_method="joblib_SHA1",
            type="numpy.int64", id=id(res),
            details={'shape': (), 'dtype': np.int64}, value=3)

        expected_slice_output = DataObject(
            hash=joblib_hash(TEST_ARRAY[:2]), hash_method="joblib_SHA1",
            type="numpy.ndarray", id=id(TEST_ARRAY[:2]),
            details={'shape': (2,), 'dtype': np.int64}, value=None)

        _check_function_execution(
            actual=Provenance.history[0],
            exp_function=FunctionInfo('subscript', '', ''),
            exp_input={0: TEST_ARRAY_INFO},
            exp_params={'slice': ':2'},
            exp_output={0: expected_slice_output},
            exp_arg_map=None, exp_kwarg_map=None, exp_code_stmnt=None,
            exp_return_targets=[], exp_order=None,
            test_case=self)

        _check_function_execution(
            actual=Provenance.history[1],
            exp_function=FunctionInfo('add_numbers_array',
                                      'test_code_analysis', ''),
            exp_input={'array': expected_slice_output},
            exp_params={},
            exp_output={0: expected_output},
            exp_arg_map=['array'],
            exp_kwarg_map=[],
            exp_code_stmnt="res = add_numbers_array(source_array[:2])",
            exp_return_targets=['res'],
            exp_order=1,
            test_case=self)

    def test_subscript_slice_no_stop(self):
        activate(clear=True)
        source_array = TEST_ARRAY
        res = add_numbers_array(source_array[1:])
        deactivate()

        self.assertEqual(len(Provenance.history), 2)

        expected_output = DataObject(
            hash=joblib_hash(TEST_ARRAY[1]+TEST_ARRAY[2]),
            hash_method="joblib_SHA1",
            type="numpy.int64", id=id(res),
            details={'shape': (), 'dtype': np.int64}, value=5)

        expected_slice_output = DataObject(
            hash=joblib_hash(TEST_ARRAY[1:]), hash_method="joblib_SHA1",
            type="numpy.ndarray", id=id(TEST_ARRAY[1:]),
            details={'shape': (2,), 'dtype': np.int64}, value=None)

        _check_function_execution(
            actual=Provenance.history[0],
            exp_function=FunctionInfo('subscript', '', ''),
            exp_input={0: TEST_ARRAY_INFO},
            exp_params={'slice': '1:'},
            exp_output={0: expected_slice_output},
            exp_arg_map=None, exp_kwarg_map=None, exp_code_stmnt=None,
            exp_return_targets=[], exp_order=None,
            test_case=self)

        _check_function_execution(
            actual=Provenance.history[1],
            exp_function=FunctionInfo('add_numbers_array',
                                      'test_code_analysis', ''),
            exp_input={'array': expected_slice_output},
            exp_params={},
            exp_output={0: expected_output},
            exp_arg_map=['array'],
            exp_kwarg_map=[],
            exp_code_stmnt="res = add_numbers_array(source_array[1:])",
            exp_return_targets=['res'],
            exp_order=1,
            test_case=self)

    def test_subscript_slice_step(self):
        activate(clear=True)
        source_array = TEST_ARRAY
        res = add_numbers_array(source_array[::2])
        deactivate()

        self.assertEqual(len(Provenance.history), 2)

        expected_output = DataObject(
            hash=joblib_hash(TEST_ARRAY[0]+TEST_ARRAY[2]),
            hash_method="joblib_SHA1",
            type="numpy.int64", id=id(res),
            details={'shape': (), 'dtype': np.int64}, value=4)

        expected_slice_output = DataObject(
            hash=joblib_hash(TEST_ARRAY[::2]), hash_method="joblib_SHA1",
            type="numpy.ndarray", id=id(TEST_ARRAY[::2]),
            details={'shape': (2,), 'dtype': np.int64}, value=None)

        _check_function_execution(
            actual=Provenance.history[0],
            exp_function=FunctionInfo('subscript', '', ''),
            exp_input={0: TEST_ARRAY_INFO},
            exp_params={'slice': '::2'},
            exp_output={0: expected_slice_output},
            exp_arg_map=None, exp_kwarg_map=None, exp_code_stmnt=None,
            exp_return_targets=[], exp_order=None,
            test_case=self)

        _check_function_execution(
            actual=Provenance.history[1],
            exp_function=FunctionInfo('add_numbers_array',
                                      'test_code_analysis', ''),
            exp_input={'array': expected_slice_output},
            exp_params={},
            exp_output={0: expected_output},
            exp_arg_map=['array'],
            exp_kwarg_map=[],
            exp_code_stmnt="res = add_numbers_array(source_array[::2])",
            exp_return_targets=['res'],
            exp_order=1,
            test_case=self)

    def test_subscript_index_str(self):
        activate(clear=True)
        source_dict = TEST_DICT
        res = add_numbers_array(source_dict['numbers'])
        deactivate()

        self.assertEqual(len(Provenance.history), 2)

        expected_output = DataObject(
            hash=joblib_hash(np.sum(TEST_ARRAY)),
            hash_method="joblib_SHA1",
            type="numpy.int64", id=id(res),
            details={'shape': (), 'dtype': np.int64}, value=6)

        _check_function_execution(
            actual=Provenance.history[0],
            exp_function=FunctionInfo('subscript', '', ''),
            exp_input={0: TEST_DICT_INFO},
            exp_params={'index': 'numbers'},
            exp_output={0: TEST_ARRAY_INFO},
            exp_arg_map=None, exp_kwarg_map=None, exp_code_stmnt=None,
            exp_return_targets=[], exp_order=None,
            test_case=self)

        _check_function_execution(
            actual=Provenance.history[1],
            exp_function=FunctionInfo('add_numbers_array',
                                      'test_code_analysis', ''),
            exp_input={'array': TEST_ARRAY_INFO},
            exp_params={},
            exp_output={0: expected_output},
            exp_arg_map=['array'],
            exp_kwarg_map=[],
            exp_code_stmnt="res = add_numbers_array(source_dict['numbers'])",
            exp_return_targets=['res'],
            exp_order=1,
            test_case=self)

    def test_subscript_index_from_variable(self):
        activate(clear=True)
        source_dict = TEST_DICT
        param_name = 'numbers'
        res = add_numbers_array(source_dict[param_name])
        deactivate()

        self.assertEqual(len(Provenance.history), 2)

        expected_output = DataObject(
            hash=joblib_hash(np.sum(TEST_ARRAY)),
            hash_method="joblib_SHA1",
            type="numpy.int64", id=id(res),
            details={'shape': (), 'dtype': np.int64}, value=6)

        _check_function_execution(
            actual=Provenance.history[0],
            exp_function=FunctionInfo('subscript', '', ''),
            exp_input={0: TEST_DICT_INFO},
            exp_params={'index': 'numbers'},
            exp_output={0: TEST_ARRAY_INFO},
            exp_arg_map=None, exp_kwarg_map=None, exp_code_stmnt=None,
            exp_return_targets=[], exp_order=None,
            test_case=self)

        _check_function_execution(
            actual=Provenance.history[1],
            exp_function=FunctionInfo('add_numbers_array',
                                      'test_code_analysis', ''),
            exp_input={'array': TEST_ARRAY_INFO},
            exp_params={},
            exp_output={0: expected_output},
            exp_arg_map=['array'],
            exp_kwarg_map=[],
            exp_code_stmnt="res = add_numbers_array(source_dict[param_name])",
            exp_return_targets=['res'],
            exp_order=1,
            test_case=self)

    def test_attribute(self):
        activate(clear=True)
        container_of_array = ContainerOfArray(TEST_ARRAY)
        res = add_numbers_array(container_of_array.array)
        deactivate()

        self.assertEqual(len(Provenance.history), 2)

        expected_output = DataObject(
            hash=joblib_hash(np.sum(TEST_ARRAY)),
            hash_method="joblib_SHA1",
            type="numpy.int64", id=id(res),
            details={'shape': (), 'dtype': np.int64}, value=6)

        expected_container_info = DataObject(
            hash=joblib_hash(container_of_array), hash_method="joblib_SHA1",
            type="test_code_analysis.ContainerOfArray",
            id=id(container_of_array), details={'array': TEST_ARRAY},
            value=None)

        _check_function_execution(
            actual=Provenance.history[0],
            exp_function=FunctionInfo('attribute', '', ''),
            exp_input={0: expected_container_info},
            exp_params={'name': 'array'},
            exp_output={0: TEST_ARRAY_INFO},
            exp_arg_map=None, exp_kwarg_map=None, exp_code_stmnt=None,
            exp_return_targets=[], exp_order=None,
            test_case=self)

        _check_function_execution(
            actual=Provenance.history[1],
            exp_function=FunctionInfo('add_numbers_array',
                                      'test_code_analysis', ''),
            exp_input={'array': TEST_ARRAY_INFO},
            exp_params={},
            exp_output={0: expected_output},
            exp_arg_map=['array'],
            exp_kwarg_map=[],
            exp_code_stmnt="res = add_numbers_array(container_of_array.array)",
            exp_return_targets=['res'],
            exp_order=1,
            test_case=self)

    def test_attribute_method_call(self):
        activate(clear=True)
        object_with_method = ObjectWithMethod()
        container_of_array = ContainerOfArray(TEST_ARRAY)
        res = object_with_method.add_numbers(container_of_array.array)
        deactivate()

        self.assertEqual(len(Provenance.history), 2)

        expected_output = DataObject(
            hash=joblib_hash(np.sum(TEST_ARRAY)),
            hash_method="joblib_SHA1",
            type="numpy.int64", id=id(res),
            details={'shape': (), 'dtype': np.int64},
            value=6)

        object_info = DataObject(
            hash=joblib_hash(object_with_method),
            hash_method="joblib_SHA1",
            type="test_code_analysis.ObjectWithMethod",
            id=id(object_with_method),
            details={},
            value=None)

        expected_container_info = DataObject(
            hash=joblib_hash(container_of_array), hash_method="joblib_SHA1",
            type="test_code_analysis.ContainerOfArray",
            id=id(container_of_array), details={'array': TEST_ARRAY},
            value=None)

        _check_function_execution(
            actual=Provenance.history[0],
            exp_function=FunctionInfo('attribute', '', ''),
            exp_input={0: expected_container_info},
            exp_params={'name': 'array'},
            exp_output={0: TEST_ARRAY_INFO},
            exp_arg_map=None, exp_kwarg_map=None, exp_code_stmnt=None,
            exp_return_targets=[], exp_order=None,
            test_case=self)

        _check_function_execution(
            actual=Provenance.history[1],
            exp_function=FunctionInfo('ObjectWithMethod.add_numbers',
                                      'test_code_analysis', ''),
            exp_input={'self': object_info, 'array': TEST_ARRAY_INFO},
            exp_params={},
            exp_output={0: expected_output},
            exp_arg_map=['self', 'array'],
            exp_kwarg_map=[],
            exp_code_stmnt="res = object_with_method.add_numbers(container_of_array.array)",
            exp_return_targets=['res'],
            exp_order=1,
            test_case=self)

    def test_subscript_initializer(self):
        activate(clear=True)
        list_1 = [1, 2, 3]
        list_2 = [3, 4, 5]
        source_data = [list_1, list_2]
        custom_object = CustomObject(source_data[0])
        deactivate()

        self.assertEqual(len(Provenance.history), 2)

        expected_output = DataObject(
            hash=joblib_hash(custom_object),
            hash_method="joblib_SHA1",
            type="test_code_analysis.CustomObject", id=id(custom_object),
            details={'data': list_1}, value=None)

        source_list_info = DataObject(
            hash=joblib_hash(source_data),
            hash_method="joblib_SHA1",
            type="builtins.list", id=id(source_data), details={},
            value=None)

        element_info = DataObject(
            hash=joblib_hash(list_1),
            hash_method="joblib_SHA1",
            type="builtins.list", id=id(list_1), details={},
            value=None)

        _check_function_execution(
            actual=Provenance.history[0],
            exp_function=FunctionInfo('subscript', '', ''),
            exp_input={0: source_list_info},
            exp_params={'index': 0},
            exp_output={0: element_info},
            exp_arg_map=None, exp_kwarg_map=None, exp_code_stmnt=None,
            exp_return_targets=[], exp_order=None,
            test_case=self)

        _check_function_execution(
            actual=Provenance.history[1],
            exp_function=FunctionInfo('CustomObject.__init__',
                                      'test_code_analysis', ''),
            exp_input={'data': element_info},
            exp_params={},
            exp_output={0: expected_output},
            exp_arg_map=['self', 'data'],
            exp_kwarg_map=[],
            exp_code_stmnt="custom_object = CustomObject(source_data[0])",
            exp_return_targets=['custom_object'],
            exp_order=1,
            test_case=self)


if __name__ == "__main__":
    unittest.main()



import unittest

from alpaca.utils.files import _get_prov_file_format, _get_file_format
from alpaca.utils import get_file_name
from os.path import expanduser


USER_PATH = expanduser("~")


class FileUtilsTestCase(unittest.TestCase):

    def test_get_file_name_user_path_new_extension(self):
        source_path = "~/test.py"
        new_path = get_file_name(source_path, None, "ttl")
        self.assertEqual(new_path, USER_PATH + "/test.ttl")

    def test_get_file_name_file_suffix_new_extension(self):
        source_path = "~/test.py"
        new_path = get_file_name(source_path, None, "ttl",
                                 suffix="_5")
        self.assertEqual(new_path, USER_PATH + "/test_5.ttl")

    def test_get_file_name_file_suffix(self):
        source_path = "~/test.py"
        new_path = get_file_name(source_path, None, None,
                                 suffix="_5")
        self.assertEqual(new_path, USER_PATH + "/test_5.py")

    def test_get_file_name_user_path_new_extension2(self):
        source_path = "~/test.py"
        new_path = get_file_name(source_path, None, ".ttl")
        self.assertEqual(new_path, USER_PATH + "/test.ttl")

    def test_get_file_name_abs_path_new_extension(self):
        source_path = "/home/user/test.py"
        new_path = get_file_name(source_path, None, "ttl")
        self.assertEqual(new_path, "/home/user/test.ttl")

    def test_get_file_name_abs_path_new_extension2(self):
        source_path = "/home/user/test.py"
        new_path = get_file_name(source_path, None, ".ttl")
        self.assertEqual(new_path, "/home/user/test.ttl")

    def test_get_file_name_new_path_same_extension(self):
        source_path = "/home/user/test.py"
        new_path = get_file_name(source_path, "/user_result", None)
        self.assertEqual(new_path, "/user_result/test.py")

    def test_get_file_name_new_path_new_extension(self):
        source_path = "/home/user/test.py"
        new_path = get_file_name(source_path, "/user_result", ".rdf")
        self.assertEqual(new_path, "/user_result/test.rdf")

    def test_get_file_name_relative_path(self):
        source_path = "/home/../user/test.py"
        new_path = get_file_name(source_path, None, ".rdf")
        self.assertEqual(new_path, "/user/test.rdf")

    def test_get_file_name_new_relative_path(self):
        source_path = "/home/user/test.py"
        new_path = get_file_name(source_path, "/test/../", None)
        self.assertEqual(new_path, "/test.py")

    def test_get_file_name_no_change_user(self):
        source_path = "~/test.py"
        new_path = get_file_name(source_path, None, None)
        self.assertEqual(new_path, USER_PATH + "/test.py")

    def test_get_file_name_no_change_relative(self):
        source_path = "/home/user/../test.py"
        new_path = get_file_name(source_path, None, None)
        self.assertEqual(new_path, "/home/test.py")

    def test_get_file_name_no_change_abs(self):
        source_path = "/home/test.py"
        new_path = get_file_name(source_path, None, None)
        self.assertEqual(new_path, "/home/test.py")

    def test_get_prov_file_format_invalid(self):
        file_name = "/test_file"
        self.assertIsNone(_get_prov_file_format(file_name))

    def test_get_prov_file_format_ttl(self):
        file_name = "/test_file.ttl"
        self.assertEqual(_get_prov_file_format(file_name), "turtle")

    def test_get_prov_file_format_rdf(self):
        file_name = "/test_file.rdf"
        self.assertEqual(_get_prov_file_format(file_name), "xml")

    def test_get_prov_file_format_json(self):
        file_name = "/test_file.json"
        self.assertEqual(_get_prov_file_format(file_name), "json-ld")

    def test_get_prov_file_format_n3(self):
        file_name = "/test_file.n3"
        self.assertEqual(_get_prov_file_format(file_name), "n3")

    def test_get_prov_file_format_other(self):
        file_name = "/test_file.other"
        self.assertEqual(_get_prov_file_format(file_name), "other")

    def test_get_file_format_no_ext(self):
        file_name = "/home/test"
        self.assertEqual(_get_file_format(file_name), None)

    def test_get_file_format(self):
        file_name = "/home/test.png"
        self.assertEqual(_get_file_format(file_name), "png")


if __name__ == "__main__":
    unittest.main()



"""
For fine control of the provenance tracking with Alpaca, some settings can be
set globally.

Every time a tracked function runs, the current state of the
setting is used. The elements in the provenance records produced retain
information describing which setting value was used, when applicable.

Currently, the following settings can be defined:

* **use_builtin_hash_for_module**: list of str
        Objects from the packages defined in the list will be hashed using
        the builtin `hash` function, instead of `joblib.hash`.

        Alpaca uses the `joblib.hash` function with SHA1 algorithm to obtain
        hashes used to identify the objects. This may be problematic for some
        packages, as the objects may be subject to changes that are not useful
        to be tracked in detail.

        One example is `matplotlib.Axes` objects. `joblib.hash` is sensitive
        enough to produce a different hash every time the object is modified
        (e.g., adding a new plot, legend, etc.). For `matplotlib.Axes`
        objects, the builtin `hash` function is able to disambiguate each
        object instance among all other `matplotlib.Axes` objects within
        the script scope, but every modification will not produce a change in
        the hash value. Therefore, if adding `'matplotlib'` to the list in this
        global setting, the chain of changes in `matplotlib.Axes` objects are
        not going to be shown and the provenance track will be simplified.

        If objects of the package are elements of a container (e.g., list or
        NumPy array containing `matplotlib.Axes` objects) a special hash will
        be computed for the container, using the builtin `hash` (i.e., hash of
        the tuple containing the hashes of each element, which is obtained
        using the builtin `hash`).

        Default: []

* **authority**: str
        The string defining the authority component used in the identifiers
        throughout Alpaca.

        Data objects, files, scripts, functions, and function executions are
        serialized to RDF using URN identifiers. The basic form of the
        identifier string is `urn:[authority]:alpaca:[complement]`, where
        `[complement]` is a string composed by specific information of each
        element identified.

        `[authority]` is a string that points to the institute or organisation
        which has responsibility over the script execution, and is defined
        by this setting.

        Default: "my-authority"

* **store_values**: list of str
        The values of the objects from the types in the list will be stored
        together with the provenance information. Note that objects of the
        builtin types `str`, `bool`, `int`, `float` and `complex`, as well as
        the NumPy numeric types (e.g. `numpy.float64`) are stored by default.
        This option should be used to store values of more complex types, such
        as dictionaries. In this case, the list in this setting should have
        the `builtins.dict` entry. The strings are the full path to the Python
        object, i.e., `[module].[...].[object_class]`.


To set/read a setting, use the function :func:`alpaca_setting`.

.. autofunction :: alpaca.alpaca_setting
"""

# Global Alpaca settings dictionary
# Should be modified only through the `alpaca_setting` function.

_ALPACA_SETTINGS = {'use_builtin_hash_for_module': [],
                    'authority': "my-authority",
                    'store_values': []}


def alpaca_setting(name, value=None):
    """ Gets/sets a global Alpaca setting.

    Parameters
    ----------
    name : str
        Name of the setting.
    value : Any, optional
        If not None, the setting `name` will be defined with `value`.
        If None, the current value of setting `name` will be returned.
        Default: None

    Returns
    -------
        The new value of setting `name` or its current value.

    Raises
    ------
    ValueError
        If `name` is not one of the global settings used by Alpaca or if
        the type of `value` is not compatible. Check the documentation for
        valid names and their description.
    """
    if name not in _ALPACA_SETTINGS:
        raise ValueError(f"Setting '{name}' is not valid.")

    if value is not None:
        expected_type = type(_ALPACA_SETTINGS[name])
        if type(value) is not expected_type:
            raise ValueError(f"Setting '{name}' must be '{expected_type}'")
        _ALPACA_SETTINGS[name] = value

    return _ALPACA_SETTINGS[name]



"""
This class reads a file with serialized provenance data into a NetworkX graph.

It provides functionality for manipulating the graph to simplify the
visualization, and also to select which details of the captured information
will be displayed as node attributes. Finally, it allows saving the graph in
formats used by graph visualization software, such as GEXF or GraphML.

See the :ref:`visualization` section on the `Installation` section,
for instructions on how to download and setup Gephi that can be used to
visualize GEXF files.

.. autoclass:: alpaca.ProvenanceGraph
    :members:

"""


import re
from itertools import chain
from collections import defaultdict
import logging

import networkx as nx
from networkx.algorithms.summarization import (_snap_eligible_group,
                                               _snap_split)

from rdflib.namespace import RDF, PROV

from alpaca.ontology import ALPACA
from alpaca.serialization import AlpacaProvDocument
from alpaca.serialization.identifiers import (NSS_FUNCTION, NSS_FILE,
                                              entity_info, activity_info)
from alpaca.utils.files import _get_file_format


# Create logger and set configuration
logger = logging.getLogger(__file__)
log_handler = logging.StreamHandler()
log_handler.setFormatter(logging.Formatter("[%(asctime)s] alpaca.graph -"
                                           " %(levelname)s: %(message)s"))
logger.addHandler(log_handler)
logger.propagate = False


# String constants to use in the output
# These may be added to the names of `NameValuePair` information
PREFIX_ATTRIBUTE = "attribute"
PREFIX_ANNOTATION = "annotation"
PREFIX_PARAMETER = "parameter"


# Mapping of ontology predicates to the prefixes
ATTR_NAMES = {ALPACA.hasAttribute: PREFIX_ATTRIBUTE,
              ALPACA.hasAnnotation: PREFIX_ANNOTATION,
              ALPACA.hasParameter: PREFIX_PARAMETER}


def _add_gephi_interval(data, order):
    if not "gephi_interval" in data:
        data["gephi_interval"] = []
    data["gephi_interval"].append((order, order))


def _get_function_call_data(activity, function_name, execution_order, params,
                            use_name_in_parameter=True,
                            use_class_in_name=True):

    data = activity_info(activity)
    label = function_name
    if not use_class_in_name:
        label = function_name.split(".")[-1]
    data['label'] = label
    data["execution_order"] = execution_order

    prefix = "parameter" if not use_name_in_parameter else data["label"]
    for param, value in params.items():
        data[f"{prefix}:{param}"] = value

    return data


def _add_attribute(data, attr_name, attr_type, attr_value, strip_namespace):
    if not strip_namespace:
        attr_name = f"{ATTR_NAMES[attr_type]}:{attr_name}"

    if attr_name in data:
        raise ValueError(
            "Duplicate property values. Make sure to include the namespaces!")
    data[attr_name] = attr_value


def _get_name_value_pair(graph, bnode):
    # Read name and value from the NameValuePair blank node
    attr_name = str(list(graph.objects(bnode, ALPACA.pairName))[0])
    attr_value = str(list(graph.objects(bnode, ALPACA.pairValue))[0])
    return attr_name, attr_value


def _get_entity_data(graph, entity, annotations=None, attributes=None,
                     strip_namespace=True, value_attribute=None):
    filter_map = defaultdict(list)

    filter_map.update(
        {ALPACA.hasAnnotation: annotations if annotations else [],
         ALPACA.hasAttribute: attributes if attributes else []})

    data = entity_info(entity)

    if annotations or attributes:
        for attr_type in (ALPACA.hasAttribute, ALPACA.hasAnnotation):
            for name_value_bnode in graph.objects(entity, attr_type):
                attr_name, attr_value = _get_name_value_pair(graph,
                                                             name_value_bnode)
                if (attr_name in filter_map[attr_type]) or \
                        filter_map[attr_type] == 'all':

                    _add_attribute(data, attr_name, attr_type, attr_value,
                                   strip_namespace)

    # Get the stored value if requested and present
    if value_attribute:
        value = graph.value(entity, PROV.value)
        if value:
            data[value_attribute] = value.toPython()

    if data['type'] == NSS_FILE:
        file_path = str(list(graph.objects(entity, ALPACA.filePath))[0])
        data["File_path"] = file_path

    return data


# Main graph class

class ProvenanceGraph:
    """
    Directed Acyclic Graph representing the provenance history stored in an
    RDF file structured with the Alpaca ontology.

    The visualization is based on NetworkX, and the graph can be accessed
    through the :attr:`graph` attribute.

    `DataObjectEntity` and `FileEntity` individuals are nodes, identified with
    the respective URIs. `FunctionExecution` activities are also loaded as
    nodes. Each of the three node types is identified by the `type` node
    attribute. Interval strings for timeline visualization in Gephi are
    provided as the `Time Interval` node attribute.

    Each node has an attributes dictionary with general description:

    * for `DataObjectEntity`, the `label` node attribute contains the Python
      class name of the object (e.g., `ndarray`). The `Python_name` node
      attribute contains the full path to the class in the package (e.g.,
      `numpy.ndarray`);
    * for `FileEntity`, the `label` node attribute is `File`;
    * for `FunctionExecution` activities, the `label` will be the function
      name (e.g. `mean`), and the `Python_name` node attribute will be the
      full path to the function in the package (e.g., `numpy.mean`).

    Each node may also have additional attributes in the dictionary, with
    extended information:

    * for `DataObjectEntity`, it contains the Python object attributes and
      annotations that were saved as metadata in the PROV file;
    * for `FileEntity`, it contains the file information such as path and hash;
    * for `FunctionExecution` activities, it contains the values of the
      parameters used to call the function.

    The node attributes to be included are selected by the `annotations` and
    `attributes` parameters during the initialization.

    Finally, the graph can be simplified using methods for condensing
    memberships (e.g., elements inside lists) and simplification (e.g.,
    repeated operation in tracks generated from loops).

    Parameters
    ----------
    prov_file : str or Path-like
        Source file(s) with RDF provenance data in the Alpaca format based on
        W3C PROV-O. If multiple files are provided, all will be loaded into
        the same graph object. This is useful to integrate provenance captured
        from several sources for visualization (e.g., steps in workflows
        or parallel processes).
    annotations : tuple of str or 'all', optional
        Names of all annotations of the objects to display in the graph as
        node attributes. Annotations are defined as values of an annotation
        dictionary that might be present in the object (e.g., Neo objects).
        In the PROV file, they are identified with the `hasAnnotation`
        property in individuals of the `DataObjectEntity` class. If `'all'`,
        all the annotations in the objects are going to be included.
        Default: None
    attributes : tuple of str or 'all', optional
        Names of all attributes of the objects to display in the graph as
        node attributes. Attributes are regular Python object attributes.
        In the PROV file, they are identified with the `hasAttribute`
        property in individuals of the `DataObjectEntity` class.  If `'all'`,
        all the attributes in the objects are going to be included.
        Default: None
    strip_namespace : bool, optional
        If False, the namespaces (i.e., `attribute` or `annotation`) will
        be shown for each requested attribute/annotation. For example, for an
        attribute `'shape'`, if `strip_namespace` is False, the key in the node
        attributes will be the full name `'attribute:shape'`. If True, the key
        in the node attributes will be just `'shape'`. The namespaces are
        `annotation` and `attribute` for object annotations and attributes,
        respectively.
        Default: True
    remove_none : bool, optional
        If True, the return nodes of functions that return `None` will be
        removed from the graph. This is useful to avoid cluttering if a
        function that returns None is called frequently.
        Default: True
    use_name_in_parameter : bool, optional
        If True, the function name will be added to the parameter name in the
        node attributes (e.g., `'function:param'`). If False, the parameter
        name will be shown with a generic tag (e.g., `'parameter:param'`). Use
        this option if different functions share same parameter names, to avoid
        ambiguity.
        Default: True
    use_class_in_method_name : bool, optional
        If True, function nodes that are methods in classes will be labeled
        with the class name as prefix (e.g., `ClassName.method_name`). If
        False, only the method name will appear in the node label (e.g.,
        `method_name`).
        Default: True
    time_intervals : bool, optional
        If True, the nodes will have the `Time Interval` attribute containing
        time interval strings in the format supported by the Gephi timeline
        feature. If False, the attribute is not included.
        Default: True
    value_attribute : str, optional
        If provided, an attribute named `value_attribute` will be added to
        the node attributes to show the values stored in the provenance
        information. Alpaca stores the values of objects of the builtin types
        `str`, `bool`, `int`, `float` and `complex`, as well as the NumPy
        numeric types (e.g. `numpy.float64`) by default. The values of
        additional types can be defined using the
        :func:`alpaca.settings.alpaca_setting` function.
        Default: None

    Attributes
    ----------
    graph : nx.DiGraph
        The NetworkX graph object representing the provenance read from the
        PROV file.

    """

    def __init__(self, *prov_file, annotations=None, attributes=None,
                 strip_namespace=True, remove_none=True,
                 use_name_in_parameter=True, use_class_in_method_name=True,
                 time_intervals=True, value_attribute=None):

        # Load PROV records from the file(s)
        doc = AlpacaProvDocument()
        for file in prov_file:
            doc.read_records(file, file_format=None)

        # Transform RDFlib graph to NetworkX and simplify the graph for
        # visualization. The parameters passed to the class initialization
        # will control the graph output
        self.graph = self._transform_graph(
            doc.graph, annotations=annotations, attributes=attributes,
            strip_namespace=strip_namespace, remove_none=remove_none,
            use_name_in_parameter=use_name_in_parameter,
            use_class_in_method_name=use_class_in_method_name,
            time_intervals=time_intervals, value_attribute=value_attribute
        )

        if time_intervals:
            # Nodes that are not directly connected to function call nodes,
            # need to have the execution counter set, so that the Gephi
            # timeline visualization works
            self._find_missing_intervals(self.graph)

            # Generate the interval according to Gephi format
            self._generate_interval_strings(self.graph)

    @staticmethod
    def _find_missing_intervals(graph):
        # Find all membership nodes and create a subgraph with the ancestors
        # and successors.
        # We will have all the spots where the execution counter was not set.
        # Then we build, for each path from the bottom to the top, the full
        # list with intervals at each node.

        processed_nodes = []
        subgraph_nodes = []
        for u, v, data in graph.edges(data=True):
            if data['membership']:
                subgraph_nodes.extend([u, v])

        subgraph = graph.subgraph(subgraph_nodes).reverse(copy=True)

        # We do progressively, changing the successors of the root nodes
        # each time, and generating a new subgraph until no nodes remain
        while not nx.is_empty(subgraph):
            root_nodes = [node for node in subgraph.nodes if
                          subgraph.in_degree(node) == 0]
            for root in root_nodes:
                successors = subgraph.successors(root)
                interval = subgraph.nodes[root]["gephi_interval"]
                for succ in successors:
                    attrs = graph.nodes[succ]
                    if not "gephi_interval" in attrs:
                        attrs["gephi_interval"] = []
                    attrs["gephi_interval"].extend(interval)

            processed_nodes.extend(root_nodes)
            subgraph_nodes = []
            for u, v, data in graph.edges(data=True):
                if data['membership'] and v not in processed_nodes:
                    subgraph_nodes.extend([u, v])

            subgraph = graph.subgraph(subgraph_nodes).reverse(copy=True)

    @staticmethod
    def _generate_interval_strings(graph):
        # Create a Gephi interval string (e.g. "<[start:end];[start:end]>")
        # for each node in `graph`, and add as additional node data, using
        # the "Time Interval" label.

        for node, data in graph.nodes(data=True):
            data["gephi_interval"].sort(key=lambda tup: tup[0])
            segments = ";".join([f"[{start:.1f},{stop:.1f}]" for start, stop in
                                 data["gephi_interval"]])
            interval = f"<{segments}>"
            data["Time Interval"] = interval
            data.pop("gephi_interval")

    @staticmethod
    def _transform_graph(graph, annotations=None, attributes=None,
                         strip_namespace=True, remove_none=True,
                         use_name_in_parameter=True,
                         use_class_in_method_name=True,
                         time_intervals=True, value_attribute=None):
        # Transform an RDFlib graph obtained from the PROV data, so that the
        # visualization is simplified. A new `nx.DiGraph` object is created
        # and returned. Annotations and attributes of the entities stored in
        # the PROV file can be filtered.

        transformed = nx.DiGraph()
        none_nodes = []

        logger.debug("Creating nodes")

        # Copy all the Entity nodes, while adding the requested attributes and
        # annotations as node data.
        for entity in chain(graph.subjects(RDF.type, ALPACA.DataObjectEntity),
                            graph.subjects(RDF.type, ALPACA.FileEntity)):
            node_id = str(entity)
            if remove_none and "builtins.NoneType" in node_id:
                none_nodes.append(node_id)
                continue
            data = _get_entity_data(graph, entity,
                                    annotations=annotations,
                                    attributes=attributes,
                                    strip_namespace=strip_namespace,
                                    value_attribute=value_attribute)
            transformed.add_node(node_id, **data)

        # Add all the edges.
        # If usage/generation, create additional nodes for the function call,
        # with the parameters as node data.
        # If membership, membership flag is set to True, as this will be used.
        logger.debug("Creating edges")

        for s, func_execution in graph.subject_objects(PROV.wasGeneratedBy):

            target = str(s)

            # Extract all the parameters of the function execution
            params = dict()
            for parameter in graph.objects(func_execution,
                                           ALPACA.hasParameter):
                name, value = _get_name_value_pair(graph, parameter)
                params[name] = value

            # Execution order
            execution_order = list(
                graph.objects(func_execution,
                              ALPACA.executionOrder))[0].value

            # Function description
            function = list(
                graph.objects(func_execution, ALPACA.usedFunction))[0]
            function_name = list(
                graph.objects(function, ALPACA.functionName))[0].value

            # Get the entity(ies) used for this generation
            source_entities = list()
            for entity in graph.objects(func_execution, PROV.used):
                source_entities.append(str(entity))

            node_data = _get_function_call_data(activity=func_execution,
                function_name=function_name,
                execution_order=execution_order, params=params,
                use_name_in_parameter=use_name_in_parameter,
                use_class_in_name=use_class_in_method_name)

            if time_intervals:
                _add_gephi_interval(node_data, node_data["execution_order"])

            # Add a new node for the function execution, with the activity
            # data
            node_id = str(func_execution)
            if not node_id in transformed.nodes:
                transformed.add_node(node_id, **node_data)

            # Add all the edges from sources to activity and from activity
            # to targets
            for source in source_entities:
                transformed.add_edge(source, node_id, membership=False)
                if time_intervals:
                    _add_gephi_interval(transformed.nodes[source],
                                    node_data['execution_order'])

            if not remove_none or (remove_none and target not in none_nodes):
                transformed.add_edge(node_id, target, membership=False)
                if time_intervals:
                    _add_gephi_interval(transformed.nodes[target],
                                    node_data['execution_order'])

        for container, member in graph.subject_objects(PROV.hadMember):

            membership_relation = None
            for predicate, object in graph.predicate_objects(member):
                if predicate in [ALPACA.containerIndex, ALPACA.containerSlice]:
                    membership_relation = f"[{str(object)}]"
                elif predicate == ALPACA.fromAttribute:
                    membership_relation = f".{str(object)}"

            if membership_relation is None:
                raise ValueError("Membership information not found for"
                                 f"{container}->{member} relation.")

            transformed.add_edge(str(container), str(member), membership=True,
                                 label=membership_relation)

        return transformed

    @staticmethod
    def _condense_memberships(graph, preserve=None):
        if preserve is None:
            preserve = []

        # Find all membership edges
        filter_edges = [tuple(e) for *e, data in graph.edges(data=True) if
                        data['membership']]

        # Iterate over the edges. We will contract if:
        #  - target does not have an edge to a function
        #  - target is not preserved

        remove_nodes = []
        replaced_edges = []

        while len(filter_edges) > 0:

            e = filter_edges.pop(0)
            if e in replaced_edges:
                continue
            u, v = e

            if graph.nodes[v]['label'] in preserve:
                continue

            successors = []
            input_to_function = False
            for successor in graph.successors(v):
                if graph.nodes[successor]['type'] == NSS_FUNCTION:
                    input_to_function = True
                    break
                successors.append(successor)
            if input_to_function:
                continue

            edge_data = graph.edges[e]

            # For each successor of node `v`, we connect with node `u`.
            # We push the replaced edges to processed edges, in case they are
            # also to be removed later.
            # Edge label is formed by concatenating the current edge with the
            # current value of `v` to the successor.
            # We add the new edge to the list to be processed, in case several
            # sequential memberships are being pruned.
            # Replaced edges are removed from the graph.
            for successor in successors:
                # Create new label
                replaced_edge = (v, successor)
                replaced_data = graph.edges[replaced_edge]
                new_edge_label = edge_data['label'] + replaced_data['label']
                replaced_data['label'] = new_edge_label

                # Create new edge
                new_edge = (u, successor)
                graph.add_edge(*new_edge, **replaced_data)
                filter_edges.append(new_edge)

                # Remove replaced edges
                graph.remove_edge(*replaced_edge)
                replaced_edges.append(replaced_edge)

            # Remove original edge
            graph.remove_edge(*e)

            if not v in remove_nodes:
                remove_nodes.append(v)

        # Remove the nodes
        for node in remove_nodes:
            graph.remove_node(node)

    def condense_memberships(self, preserve=None):
        """
        Condense sequential entity membership relationships into a single
        node. This operation is done in-place, i.e., the graph stored as
        :attr:`graph` will be modified.

        Membership relationships are used to describe relationships such as
        attributes (e.g. `block.segments`) or membership in containers (e.g.,
        `spiketrains[0]`).

        Parameters
        ----------
        preserve : tuple of str, optional
            List the labels of nodes that should not be condensed if present
            in a membership relationship.
            Default: None
        """
        self._condense_memberships(self.graph, preserve=preserve)

    @staticmethod
    def _snap_build_graph(graph, groups, neighbor_info, remove_attributes,
                          record_members=True):
        # Function modified from NetworkX 2.6, to build the aggregated graph
        # after SNAP aggregation.
        #
        # Please refer to the `Acknowledgements and open source software`
        # section for copyright and license information.

        def _aggregate_attributes(source, group_iterator):
            raw_attributes = defaultdict(set)
            for member in group_iterator:
                for attr, value in source[member].items():
                    raw_attributes[attr].add(value)

            # Transform all elements values to strings
            attributes = {
                key: str(next(iter(value)))
                if len(value) == 1 else ";".join(map(str, sorted(list(value))))
                for key, value in raw_attributes.items()
                if remove_attributes is None or key not in remove_attributes
            }

            # Organize time intervals
            if 'Time Interval' in attributes:
                intervals = re.findall(r"(\[[\d+.,]+\])",
                                       attributes['Time Interval'])
                intervals.sort()
                intervals_str = ";".join(intervals)
                attributes['Time Interval'] = f"<{intervals_str}>"

            return attributes

        output = nx.DiGraph()
        prefix = "Step"
        node_label_lookup = dict()

        for index, group_id in enumerate(groups):
            group_set = groups[group_id]
            supernode = f"{prefix} {index}"
            node_label_lookup[group_id] = supernode

            # We sumarize all possible values for all attributes in the nodes
            # from the group
            supernode_attributes = _aggregate_attributes(graph.nodes,
                                                         group_set)

            count = len(group_set)
            supernode_attributes['member_count'] = count

            # Save a string with the identifiers of all member nodes if
            # requested
            if record_members:
                supernode_attributes['members'] = ";".join(group_set)

            output.add_node(supernode, **supernode_attributes)

        for group_id in groups:
            group_set = groups[group_id]
            source_supernode = node_label_lookup[group_id]
            for other_group, group_edge_types in neighbor_info[
                next(iter(group_set))
            ].items():
                if group_edge_types:
                    target_supernode = node_label_lookup[other_group]
                    summary_graph_edge = (source_supernode, target_supernode)

                    superedge_attributes = {}
                    output.add_edge(*summary_graph_edge,
                                    **superedge_attributes)

        return output

    def aggregate(self, group_node_attributes, use_function_parameters=True,
                  output_file=None, remove_attributes=None,
                  record_members=True):
        """
        Creates a summary graph based on a selection of attributes of the
        nodes in the graph.

        The attributes can be individualized for each
        node label (as defined by the `label` node attribute), so that
        different levels of aggregation are possible. Therefore, it is
        possible to generate visualizations with different levels of detail
        to progressively inspect the provenance trace.

        In the summarized nodes, the `member_count` node attribute stores the
        number of nodes in the group. If requested, the list with the IDs of
        the original nodes that are part of that group can be stored in the
        `members` node attribute.

        Parameters
        ----------
        group_node_attributes : dict
            Dictionary selecting which attributes are used in the aggregation.
            The keys are the possible labels in the graph, and the values
            are tuples of the node attributes or callables used for
            determining supernodes.

            For example, to aggregate `Quantity` nodes based on different
            `shape` attribute values, `group_node_attributes` would be
            `{'Quantity': ('shape',)}`. If passing an empty dictionary, no
            attributes will be considered, and the aggregation will be based
            on the topology (i.e., nodes at similar levels will be grouped
            according to the connectivity).

            In addition to attribute names, callables that take the
            arguments `(graph, node, data)`, where `graph` is the graph being
            aggregated, `node` is the node being evaluated for grouping, and
            `data` is the dictionary of attributes, can be used. The returned
            value is used to define the group. This allows flexibility when
            grouping, as attribute values can be transformed (e.g., extracting
            a token such as file extension from an attribute that stores the
            path as a string), or the relationship of the node to neighbors
            and values of edges can be checked. However, this will increase
            the time to evaluate the grouping criteria of a node.
        use_function_parameters : bool, optional
            If True, the parameters of function nodes in the graph will be
            considered in the aggregation, i.e., if the same function is called
            with different parameters, different supernodes will be generated.
            If False, a single supernode will be produced, regardless of the
            different parameters used.
            Default: True
        output_file : str or Path-like, optional
            If None, a `nx.DiGraph` object will be returned. If not None, the
            graph will be saved in the provided path, and the function will
            return None. The file must have either the `.gexf` or the
            `.graphml` extension, to save as either GEXF or GraphML formats
            respectively.
            Default: None
        remove_attributes : str or tuple of str, optional
            Remove the specified node attributes from the aggregated graph.
            Default: None
        record_members : bool, optional
            If True, the summarized nodes will have the `members` attribute
            with the identifiers of all nodes that are part of the group.
            Default: True

        Returns
        -------
        nx.DiGraph or None
            If an output file was not specified in `output_file`, returns the
            aggregated graph as a NetworkX object. The original graph stored
            in :attr:`graph` is not modified.
            If an output file was specified, returns None.

        Raises
        ------
        ValueError
            If `output_file` is not None and the file does not have either
            `'.gexf'` or `'.graphml'` as extension.

        Notes
        -----
        This function is an adaptation of the `snap_aggregation` function
        included in NetworkX 2.6, which implemented the SNAP algorithm based on
        [1]_.

        The function was modified to group the nodes based on different
        attributes or callables (using a dictionary based on the labels)
        instead of attributes that are common to all nodes.

        During the summary graph generation, the attribute values are also
        summarized, so that the user has an idea of all the possible values in
        the group.

        Please refer to the :ref:`open_software_licenses` section for copyright
        and license information.

        References
        ----------
        .. [1] Y. Tian, R. A. Hankins, and J. M. Patel. Efficient aggregation
           for graph summarization. In Proc. 2008 ACM-SIGMOD Int. Conf.
           Management of Data (SIGMOD’08), pages 567–580, Vancouver, Canada,
           June 2008.
        """

        def _fetch_group_tuple(graph, node, data, label, data_attributes,
                               use_function_params):
            group_info = [label]

            # If function, we use all the parameters
            if data['type'] == NSS_FUNCTION and use_function_params:
                parameters = [name for name in data.keys()
                              if name.startswith("parameter:") or
                              name.startswith(f"{label}:")]
                parameters.sort()
                for attr in parameters:
                    group_info.append(data[attr])
            else:
                if data_attributes is not None:
                    # We have requested grouping for this object based on
                    # selected attributes/callables. Otherwise, we will use
                    # the label only

                    for attr in data_attributes:

                        if callable(attr):
                            # We have requested grouping using a function that
                            # takes the graph, the node, and the node
                            # attributes as parameters. This allows a more
                            # customized filtering, that can extract specific
                            # information from the attribute value or use the
                            # node relationships
                            group_info.append(attr(graph, node, data))
                        else:
                            # Fetch the attribute value for this node, if
                            # available
                            group_info.append(data.get(attr, None))

            return tuple(group_info)

        # We don't consider edges
        edge_types = {edge: () for edge in self.graph.edges}

        # Create the groups based on the selected conditions
        group_lookup = {
            node: _fetch_group_tuple(
                self.graph, node, attrs, attrs['label'],
                group_node_attributes.get(attrs['label'], None),
                use_function_parameters)
            for node, attrs in self.graph.nodes.items()
        }

        groups = defaultdict(set)
        for node, node_type in group_lookup.items():
            groups[node_type].add(node)

        eligible_group_id, neighbor_info = _snap_eligible_group(
            self.graph, groups, group_lookup, edge_types=edge_types)
        while eligible_group_id:
            groups = _snap_split(groups, neighbor_info, group_lookup,
                                 eligible_group_id)
            eligible_group_id, neighbor_info = _snap_eligible_group(
                self.graph, groups, group_lookup, edge_types=edge_types)

        aggregated = self._snap_build_graph(self.graph, groups, neighbor_info,
                                            remove_attributes,
                                            record_members=record_members)

        if output_file is None:
            return aggregated

        file_format = _get_file_format(output_file)
        if file_format == "gexf":
            nx.write_gexf(aggregated, output_file)
        elif file_format == "graphml":
            nx.write_graphml(aggregated, output_file)
        else:
            raise ValueError("Unknown graph format. Please provide an output"
                             "file with either '.gexf' or '.graphml'"
                             "extension")

    def remove_attributes(self, *attributes):
        """
        Remove one or more attributes from the nodes.

        Parameters
        ----------
        attributes : str
            Key(s) identifying the attribute(s) to be removed from the node
            attribute dictionary.
        """
        if len(attributes) > 1:
            for _, node_attrs in self.graph.nodes(data=True):
                for attr in attributes:
                    node_attrs.pop(attr, None)
        else:
            # Algorithm has O(N) complexity. In the case of a single attribute,
            # there is no need for a nested loop, which will improve run time.
            attr = attributes[0]
            for _, node_attrs in self.graph.nodes(data=True):
                node_attrs.pop(attr, None)

    def save_gexf(self, file_name):
        """
        Writes the current provenance graph as a GEXF file.
        """
        nx.write_gexf(self.graph, file_name)

    def save_graphml(self, file_name):
        """
        Writes the current provenance graph as a GraphML file.
        """
        nx.write_graphml(self.graph, file_name)



"""
This module implements classes for getting the relevance provenance
information from Python objects and files (e.g., hashes and metadata).
They are used by the `decorator.Provenance` class decorator to track unique
objects during the script execution, and also to represent the objects in the
PROV files generated with the Alpaca ontology (identifiers).

The classes defined in this module are not intended to be used directly by
the user, but are used internally by the `decorator.Provenance` decorator.
"""

import hashlib
import inspect
import uuid
from copy import copy
from pathlib import Path
import logging
from collections.abc import Iterable
from collections import defaultdict

import joblib
import numpy as np
from numbers import Number
from dill._dill import save_function

from alpaca.alpaca_types import DataObject, File
from alpaca.ontology.annotation import _OntologyInformation, ONTOLOGY_INFORMATION

# Need to use `dill` pickling function to support lambdas.
# Some objects may have attributes that are lambdas. One example is the
# test case of Nose. If the unit test accesses variables in the TestCase class
# (e.g., `self.spiketrains`), the hashing of the `self` object fails.
# Here we update the dispatch table of the `joblib.Hasher` object to use
# the function from `dill` that supports these lambda attributes.
joblib.hashing.Hasher.dispatch[type(save_function)] = save_function

# Create logger and set configuration
logger = logging.getLogger(__file__)
log_handler = logging.StreamHandler()
log_handler.setFormatter(logging.Formatter("[%(asctime)s]"
                                           " alpaca.data_information"
                                           " %(levelname)s: %(message)s"))
logger.addHandler(log_handler)
logger.propagate = False


class _FileInformation(object):
    """
    Class for getting information from files.

    The SHA256 hash and file path are captured.

    The method `info` is called to obtain these provenance information as the
    `File` named tuple.

    Easy comparison between files can be done using the equality operator.

    Parameters
    ----------
    file_path : str or path-like
        The path to the file that is being hashed.
    """

    @staticmethod
    def _get_content_file_hash(file_path, block_size=4096 * 1024):
        file_hash = hashlib.sha256()

        with open(file_path, 'rb') as file:
            for block in iter(lambda: file.read(block_size), b""):
                file_hash.update(block)

        return file_hash.hexdigest()

    def __init__(self, file_path):
        self.file_path = Path(file_path).expanduser().resolve().absolute()

        self._hash_type = 'sha256'
        self._hash = self._get_content_file_hash(self.file_path)

    def __eq__(self, other):
        if isinstance(other, _FileInformation):
            return self._hash == other._hash and \
                   self._hash_type == other._hash_type
        raise TypeError("Can compare only two `_FileInformation` objects")

    def __repr__(self):
        return f"{self.file_path.name}: " \
               f"[{self._hash_type}] {self._hash}"

    def info(self):
        """
        Returns provenance information for the file.

        Returns
        -------
        alpaca_types.File
            A named tuple with the following attributes:
            * hash : str
                SHA256 hash of the file.
            * hash_type: {'sha256'}
                String storing the hash type.
            * path : str or Path-like
                The path to the file that was hashed.
        """
        return File(hash=self._hash,
                    hash_type=self._hash_type,
                    path=self.file_path)


class _ObjectInformation(object):
    """
    Class for hashing Python objects and getting their information, supporting
    memoization.

    The object hash is the SHA1 hash of its value.

    The type is a string produced by the combination of the module where it
    was defined plus the type name (both returned by :func:`type`).

    Value hash is calculated using :func:`joblib.hash` or the builtin
    :func:`hash` function, depending on the `use_builtin_hash` parameter set
    during initialization.

    The values of objects of the builtin types `str`, `bool`, `int`, `complex`
    and `float` as well as the NumPy numeric types (e.g., `np.float64`) will
    be stored. Additional object types to be stored (e.g., `builtins.dict`)
    can be defined with the `store_values` parameter.

    The method `info` is called to obtain the provenance information
    associated with the object during tracking, as the `DataObject` named
    tuple. The relevant metadata attributes are also stored in the tuple.

    As the same object may be hashed several times during a single analysis
    step, a builtin memoizer stores all hashes that are computed by object ID
    (according to :func:`id`).

    Parameters
    ----------
    use_builtin_hash : list, optional
        List of package names whose object hashes will be computed using the
        Python builtin `hash` function, instead of `joblib.hash` function.
        Default: None
    store_values : list, optional
        List of object types whose values will be stored in the provenance
        information (e.g., `builtins.dict`). This is in addition to the
        builtin types `str`, `bool`, `int`, `complex` and `float` as well as
        the NumPy numeric types (e.g., `np.float64`). The values of these are
        always stored.
        Default: None
    """

    # This is a list of object attributes that provide relevant provenance
    # information. Whether the object has one defined, it will be captured.
    _metadata_attributes = ('units', 'shape', 'dtype', 't_start', 't_stop',
                            'id', 'nix_name', 'dimensionality', 'pid',
                            'create_time')

    def __init__(self, use_builtin_hash=None, store_values=None):
        self._hash_memoizer = dict()
        self._use_builtin_hash = copy(use_builtin_hash) \
            if use_builtin_hash is not None else []
        self._store_values = copy(store_values)\
            if store_values is not None else []

    @staticmethod
    def _get_object_package(obj):
        # Returns the string with the name of the package where the object
        # is defined (e.g., neo.core.SpikeTrain -> 'neo')
        module = inspect.getmodule(obj)
        package = ""
        if module is not None:
            package = module.__name__.split(".")[0]
        return package

    def _get_object_hash(self, obj, obj_type, obj_id, package):
        # Computes the hash for `obj`. `obj_type` and `package` are the
        # string representations of the type and package, and `obj_id` the
        # `id()` of the object

        logger.debug(f"{obj_type}, id={obj_id}")

        # If we already computed the hash for the object during this function
        # call, retrieve it from the memoized values
        if obj_id in self._hash_memoizer:
            return self._hash_memoizer[obj_id]

        logger.debug("Hashing")

        hash_method = None
        if package in self._use_builtin_hash:
            # Use the builtin hashing function instead of joblib's. This avoids
            # the case where object changes result in multiple object hashes,
            # which will produce a complex provenance track
            object_hash = hash(obj)
            hash_method = "Python_hash"
        else:
            # Check if the object is a container of objects that should be
            # hashed with the Python builtin function. For NumPy arrays, we
            # restrict this to arrays of dtype=object, as they are the ones
            # that can contain objects
            container_builtin_hash = False
            if isinstance(obj, Iterable) and not (
                    isinstance(obj, np.ndarray) and obj.dtype != object):

                iterator = obj if not isinstance(obj, np.ndarray) \
                    else obj.ravel()

                # Loop over elements. If any element is not from the requested
                # packages, abort the loop
                container_builtin_hash = True
                for element in iterator:
                    container_builtin_hash = \
                        container_builtin_hash and \
                        (self._get_object_package(element) in
                         self._use_builtin_hash)

                    if not container_builtin_hash:
                        break

            if container_builtin_hash:
                # We also have to use the builtin hash if objects from the
                # requested packages are stored in containers (e.g., NumPy
                # arrays with matplotlib Axes objects).

                iterator = obj if not isinstance(obj, np.ndarray) \
                    else obj.ravel()
                object_hash = joblib.hash(
                    tuple([hash(element) for element in iterator]),
                    hash_name='sha1'
                )
                hash_method = "Python_hash"
            else:
                # Other objects, like Neo, Quantity and NumPy arrays, use
                # joblib's hash function
                object_hash = joblib.hash(obj, hash_name='sha1')
                hash_method = "joblib_SHA1"

        # Memoize the hash
        self._hash_memoizer[obj_id] = (object_hash, hash_method)

        return object_hash, hash_method

    def info(self, obj):
        """
        Returns provenance information for the object, as the `DataObject`
        named tuple.

        Metadata (e.g., annotations in Neo objects) is also captured. Any
        relevant attributes are also captured as metadata.

        If the object is None, then the hash is replaced by a unique id for
        the object.

        Parameters
        ----------
        obj : object
            Python object to get the provenance information.

        Returns
        -------
        alpaca_types.DataObject
            A named tuple with the following attributes:
            * hash : str or UUID
                Hash of the object. For None objects, it will be an UUID
                generated to uniquely identify the object.
            * hash_method : {"Python_hash", "joblib_SHA1", "UUID"}
                Hash function used in the computation. If the
                :attr:`use_builtin_hash` list is defined, the builtin Python
                `hash` function is used for objects of the packages in the
                list.
                For None objects, the value will be "None".
            * type: str
                Type of the object.
            * id : int
                Reference for the object.
            * details : dict
                Extended information (metadata) on the object.
            * value : object
                For builtin objects (`str`, `int`, `float`, `bool`, `complex`)
                or equivalent objects (e.g. `numpy.float64`), the value is
                stored. Additional object types specified with the
                :attr:`store_values` list will also be stored.
        """
        type_information = type(obj)
        obj_type = f"{type_information.__module__}.{type_information.__name__}"
        obj_id = id(obj)

        # All Nones will have the same hash. Use UUID instead
        if obj is None:
            unique_id = uuid.uuid4()
            return DataObject(hash=unique_id, hash_method="UUID",
                              type=obj_type, id=obj_id, details={},
                              value=None)

        # Here we can extract specific metadata to record
        details = {}

        # Currently fetching the whole instance dictionary
        if hasattr(obj, '__dict__'):
            # Need to copy otherwise the hashes change
            details = copy(obj.__dict__)

        # Store specific attributes that are relevant for arrays, quantities
        # Neo objects, and AnalysisObjects
        for attr in self._metadata_attributes:
            if hasattr(obj, attr):
                details[attr] = getattr(obj, attr)

        # Compute object hash
        package = self._get_object_package(obj)
        obj_hash, hash_method = self._get_object_hash(obj=obj,
                                                      obj_type=obj_type,
                                                      obj_id=obj_id,
                                                      package=package)

        # Store object value
        obj_value = None
        if isinstance(obj, (str, bool, Number)):
            obj_value = obj
        elif obj_type in self._store_values:
            obj_value = str(obj)

        # Add ontology information if not present
        if (not ONTOLOGY_INFORMATION.get(obj_type) and
            _OntologyInformation.get_ontology_information(obj)):
            ONTOLOGY_INFORMATION[obj_type] = _OntologyInformation(obj)

        return DataObject(hash=obj_hash, hash_method=hash_method,
                          type=obj_type, id=obj_id, details=details,
                          value=obj_value)



"""
This module contains the OWL ontology based on the PROV-O model from W3C, that
is used to represent provenance information captured by Alpaca.
An RDFLib namespace is provided, to be used in other modules.
"""

from rdflib import Graph, Namespace
import pathlib


ONTOLOGY_SOURCE = pathlib.Path(__file__).with_name("alpaca.owl")


def get_alpaca_namespace():
    g = Graph()
    g.parse(ONTOLOGY_SOURCE, format='ttl')
    for prefix, uri in g.namespaces():
        if prefix == "":
            return Namespace(uri)
    raise ValueError("Could not get Alpaca ontology namespace.")


# Namespace object to be used with RDFLib graphs
ALPACA = get_alpaca_namespace()



"""
This module generate URN identifiers for the identification of objects within
Alpaca.
"""

import pathlib
import re

# Values that are used to compose the URNs
# URNs take the general form "urn:NID:NSS", followed by optional components
# according to RFC 8141.

NSS_FUNCTION = "function"             # Functions executed
NSS_FILE = "file"                     # Files accessed
NSS_DATA = "object"                   # Data objects (input/outputs/containers)
NSS_SCRIPT = "script"                 # The execution script
NSS_EXECUTION = "function_execution"  # Execution of a function


def get_base_urn(authority):
    return f"urn:{authority}:alpaca"


# <urn:fz-juelich.de:alpaca:object:Python:neo.core.AnalogSignal:423423432432423432432>
def data_object_identifier(object_info, authority):
    object_hash = object_info.hash
    type_string = object_info.type
    urn = f"{get_base_urn(authority)}:{NSS_DATA}:Python:{type_string}:{object_hash}"
    return urn


# <urn:fz-juelich.de:alpaca:file:sha256:234234324324324324234324>
def file_identifier(file_info, authority):
    hash_type = file_info.hash_type
    file_hash = file_info.hash
    urn = f"{get_base_urn(authority)}:{NSS_FILE}:{hash_type}:{file_hash}"
    return urn


def _get_function_name(function_info):
    function_name = ""
    if function_info.module:
        function_name = f"{function_info.module}."
    function_name = f"{function_name}{function_info.name}"
    return function_name


# <urn:fz-juelich.de:alpaca:function:Python:elephant.spectral.welch_psd>
def function_identifier(function_info, authority):
    function_name = _get_function_name(function_info)
    urn = f"{get_base_urn(authority)}:{NSS_FUNCTION}:Python:{function_name}"
    return urn


# <urn:fz-juelich.de:alpaca:script:Python:run_psd.py:f32432j34k24#4567-4567-dflsd4-dfdsfs>
def script_identifier(script_info, session_id, authority):
    script_name = pathlib.Path(script_info.path).name
    urn = f"{get_base_urn(authority)}:{NSS_SCRIPT}:Python:{script_name}:" \
          f"{script_info.hash}#{session_id}"
    return urn


def execution_identifier(script_info, function_info, session_id, execution_id,
                         authority):
    function_name = _get_function_name(function_info)
    urn = f"{get_base_urn(authority)}:{NSS_EXECUTION}:Python:" \
          f"{script_info.hash}:{session_id}:{function_name}#{execution_id}"
    return urn


# Functions to extract information from identifiers, used when generating
# visualizations with NetworkX graphs.

def _strip_local_part(identifier):
    match = re.match(r"urn:[^:]+:alpaca:(.+)", identifier)
    return match.group(1)


def entity_info(identifier):
    local_part = _strip_local_part(identifier)
    info = local_part.split(":")

    entity_type = info[0]
    data = {'type': entity_type,
            'data_hash': info[-1]}

    if entity_type == NSS_FILE:
        data['label'] = "File"
        data['hash_type'] = info[-2]
    elif entity_type == NSS_DATA:
        data['label'] = info[-2].split(".")[-1]  # label is the class name
        data['Python_name'] = info[-2]           # store full path to class

    return data


def activity_info(identifier):
    function_name, exec_id = identifier.split(":")[-1].split("#")
    data = {
        "Python_name": function_name,
        "type": NSS_FUNCTION,
        "execution_id": exec_id
    }
    return data



"""
This module defines named tuples that are used to structure the information
throughout Alpaca.
"""

from collections import namedtuple


# NAMED TUPLES TO STORE PROVENANCE INFORMATION OF EACH CALL

# In `FunctionExecution`:
#   `function`: a `FunctionInfo` named tuple;
#   `input` and `output`: dictionaries with `DataObject` or `File` named
#        tuples as values, describing the objects for the input and the
#        output of the function, respectively;
#    `arg_map`: names of the positional arguments in the function definition;
#    `kwarg_map`: names of the keyword arguments in the function definition;
#    `call_ast`: `ast.Call` node for the current function call;
#    `code_statement`: string containing the script statement that originated
#        this function call;
#    `time_stamp_start` and `time_stamp_end`: ISO datetime values as strings,
#        with the start and end times of the function execution, respectively;
#    `return_targets`: names of the variables where the outputs of the
#        function were stored;
#    `order`: integer defining the order of this function call in the whole
#        tracking history;
#    `execution_id`: UUID of this function call.

FunctionExecution = namedtuple('FunctionExecution', ('function',
                                                     'input',
                                                     'params',
                                                     'output',
                                                     'arg_map',
                                                     'kwarg_map',
                                                     'call_ast',
                                                     'code_statement',
                                                     'time_stamp_start',
                                                     'time_stamp_end',
                                                     'return_targets',
                                                     'order',
                                                     'execution_id')
                               )

FunctionInfo = namedtuple('FunctionInfo', ('name', 'module', 'version',))


# NAMED TUPLE TO STORE ARGUMENTS THAT ARE CONTAINERS

# This can be elements inside an iterator (such as a list) but also
# variable arguments that are defined in the form `*args` in the function
# definition. Each element of the `Container` named tuple will be stored as
# individual `DataObject` named tuples in `Container.elements` dictionary,
# according to their order

Container = namedtuple('Container', 'elements')


# NAMED TUPLES TO STORE HASHES AND INFORMATION ABOUT OBJECTS

# `DataObject` is for Python objects, and `File` is for files stored in
# the disk.

DataObject = namedtuple('DataObject', ('hash', 'hash_method', 'type', 'id',
                                       'details', 'value'))

File = namedtuple('File', ('hash', 'hash_type', 'path'))



"""
This module implements classes to extract and work with the information
obtained from the nodes of an Abstract Syntax Tree (AST) describing the code
associated with a given function call within the script.

Example
-------

Consider the statement below:

>>> isi_times = isi(block.segments[0].spiketrains[0])

The object that was input to the `isi` function is the first element of the
list `spiketrains`, that is an attribute from the first element of the list
`segments` from the `block` object.

The `_CallAST` object in this module takes the `Call` AST node associated with
that statement, and all these attribute/indexing operations are reconstructed,
so that the actual input to the function is described with respect to its
hierarchical membership to the `block` object. For this, the
`static_relationship_tree._StaticRelationship` objects are used.

In the end, after processing the AST `Call` node, the attribute/indexing
operations are added as individual `FunctionExecution` tuples to the
provenance history in the decorator, with the details to reconstruct the
static relationships, and with the intermediate objects added to the
provenance track.

The operations currently analyzed correspond to the `Name`, `Attribute` and
`Subscript` AST Nodes.

The classes defined in this module are not intended to be used directly by
the user, but are used internally by the `decorator.Provenance` decorator.
"""

import ast
import itertools

from alpaca.code_analysis.static_relationship_tree import (
    _AttributeRelationship, _NameRelationship, _SubscriptRelationship)


class _NameAST(ast.NodeTransformer):
    """
    NodeTransformer to find all root variables that are loaded in an
    Abstract Syntax Tree (i.e., identified by `ast.Name` nodes).

    The reference to the actual Python object is stored in the transformed
    node as the :attr:`instance`, and the object information is stored in the
    :attr:`object_info` attribute.

    Parameters
    ----------
    provenance_tracker : decorator.Provenance
        Reference to the Alpaca provenance tracker decorator.
    data_info : data_information._ObjectInformation
        Instance to the class used to hash and get information of objects in
        Alpaca.
    """

    provenance_tracker = None

    def __init__(self, provenance_tracker, data_info):
        super(_NameAST, self).__init__()
        self.provenance_tracker = provenance_tracker
        self.data_info = data_info

    def visit_Name(self, node):
        if isinstance(node.ctx, ast.Load):
            # If this is a node where the variable is accessed, get the
            # actual object from the tracked frame, and store the object
            # reference and info
            instance = self.provenance_tracker._get_script_variable(node.id)
            setattr(node, 'instance', instance)
            setattr(node, 'object_info', self.data_info.info(instance))
            return node
        return node


class _CallAST(ast.NodeVisitor):
    """
    NodeVisitor to inspect and fetch subscript/attribute relationships
    in the call to the function that is being tracked (i.e., an `ast.Call`
    node).

    This will not transform the node, only add `FunctionExecution` named
    tuples to the history.

    Parameters
    ----------
    provenance_tracker : decorator.Provenance
        Reference to the provenance tracker decorator with the history.
    data_info : data_information._ObjectInformation
        Instance to the class used to hash and get information of the objects
        in Alpaca.
    function : str
        Name of the function being tracked.
    time_stamp : datetime
        Timestamp of the current function execution.
    """

    def __init__(self, provenance_tracker, data_info, function, time_stamp):
        super(_CallAST, self).__init__()
        self.provenance_tracker = provenance_tracker
        self.function = function
        self.data_info = data_info
        self.time_stamp = time_stamp

    def visit_Call(self, node):

        # In case of initializers, the AST function name will not have
        # `__init__` in the method name
        func_name = self.function[:-9] \
            if self.function.endswith(".__init__") else self.function

        # Check if the Call is for the function being executed.
        # If a function is called using a namespace or as a method, the `func`
        # attribute will be `ast.Attribute`.
        function_in_execution = False
        if isinstance(node.func, ast.Attribute):
            attr_name = node.func.attr
            function_in_execution = func_name.endswith(f".{attr_name}")
        elif isinstance(node.func, ast.Name):
            function_in_execution = node.func.id == func_name

        if function_in_execution:
            # Fetch static information of Attribute and Subscript nodes that
            # were inputs. This should capture hierarchical information for
            # inputs that are class members or items in accessed in iterables
            for position, arg_node in enumerate(
                    itertools.chain(node.args, node.keywords)):

                if isinstance(arg_node, (ast.Subscript, ast.Attribute)):
                    _process_subscript_or_attribute(
                        node=arg_node,
                        provenance_tracker=self.provenance_tracker,
                        data_info=self.data_info,
                        time_stamp=self.time_stamp)
        else:
            # Otherwise just process the node with the generic visitor
            self.generic_visit(node)


def _fetch_object_tree(root_node, time_stamp):
    # Iterate recursively the syntax tree of `root_node`, building a
    # hierarchical tree using `_StaticRelationship` objects. This will fetch
    # the actual Python objects for every single attribute/subscript/name
    # operations.

    def _extract(node, child=None):
        if isinstance(node, ast.Subscript):
            subscript = _SubscriptRelationship(node, time_stamp, child)
            _extract(node.value, child=subscript)
            return subscript

        if isinstance(node, ast.Attribute):
            attribute = _AttributeRelationship(node, time_stamp, child)
            _extract(node.value, child=attribute)
            return attribute

        if isinstance(node, ast.Name):
            name = _NameRelationship(node, time_stamp, child)
            return name

    return _extract(root_node)


def _build_object_tree_provenance(object_tree, provenance_tracker, data_info):
    # Iterate recursively through a hierarchical tree of `_StaticRelationship`
    # objects representing the child/parent relationships between the objects,
    # and build the `FunctionExecution` tuples and store them in the history
    # of the provenance tracker decorator.

    def _get_object_info_and_store(tree_node):
        if tree_node.object_info is None:
            # Get info/hash if needed
            tree_node.object_info = data_info.info(tree_node.value)
        if tree_node.parent is not None:
            # Insert provenance information
            if tree_node.parent.object_info is None:
                # Get info/hash if needed
                tree_node.parent.object_info = data_info.info(
                    tree_node.parent.value)
            provenance_tracker.history.append(
                tree_node.get_function_execution())
            _get_object_info_and_store(tree_node.parent)

    _get_object_info_and_store(object_tree)


def _process_subscript_or_attribute(node, provenance_tracker, data_info,
                                    time_stamp):
    # From the AST starting from `node` (either `ast.Subscript` or
    # `ast.Attribute), find root variable (an `ast.Name` node), get the
    # info/hash of the associated Python object, and include reference in the
    # `ast.Name` node
    name_visitor = _NameAST(provenance_tracker, data_info)
    name_visitor.visit(node.value)

    # Fetch object references from the AST tree of `node`, and create the
    # tree of `_StaticRelationship` objects
    object_tree = _fetch_object_tree(node, time_stamp)

    # Insert provenance operations into the decorator history in
    # `provenance_tracker`, and get the remaining info/hashes if necessary
    _build_object_tree_provenance(object_tree, provenance_tracker, data_info)
